# Copyright 2017-2023 Lawrence Livermore National Security, LLC and other
# Hatchet Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

from collections import defaultdict

from .node import Node, traversal_order


def index_by(attr, objects):
    """Put objects into lists based on the value of an attribute.

    Returns:
        (dict): dictionary of lists of objects, keyed by attribute value
    """
    index = defaultdict(lambda: [])
    for obj in objects:
        index[getattr(obj, attr)].append(obj)
    return index


class Graph:
    """A possibly multi-rooted tree or graph from one input dataset."""

    def __init__(self, roots):
        assert roots is not None
        self.roots = roots

    def traverse(self, order="pre", attrs=None, visited=None):
        """Preorder traversal of all roots of this Graph.

        Arguments:
            attrs (list or str, optional): If provided, extract these
                fields from nodes while traversing and yield them. See
                :func:`~hatchet.node.traverse` for details.

        Only preorder traversal is currently supported.
        """
        # share visited dict so that we visit each node at most once.
        if visited is None:
            visited = {}

        # iterate over roots in order
        for root in sorted(self.roots, key=traversal_order):
            for value in root.traverse(order=order, attrs=attrs, visited=visited):
                yield value

    def is_tree(self):
        """True if this graph is a tree, false otherwise."""
        if len(self.roots) > 1:
            return False

        visited = {}
        list(self.traverse(visited=visited))
        return all(v == 1 for v in visited.values())

    def find_merges(self):
        """Find nodes that have the same parent and frame.

        Find nodes that have the same parent and duplicate frame, and
        return a mapping from nodes that should be eliminated to nodes
        they should be merged into.

        Return:
            (dict): dictionary from nodes to their merge targets

        """
        merges = {}  # old_node -> merged_node
        inverted_merges = defaultdict(
            lambda: []
        )  # merged_node -> list of corresponding old_nodes
        processed = []

        def _find_child_merges(node_list):
            index = index_by("frame", node_list)
            for frame, children in index.items():
                if len(children) > 1:
                    min_id = min(children, key=id)
                    for child in children:
                        prev_min = merges.get(child, min_id)
                        # Get the new merged_node
                        curr_min = min([min_id, prev_min], key=id)
                        # Save the new merged_node to the merges dict
                        # so that the merge can happen later.
                        merges[child] = curr_min
                        # Update inverted_merges to be able to set node_list
                        # to the right value.
                        inverted_merges[curr_min].append(child)

        _find_child_merges(self.roots)
        for node in self.traverse():
            if node in processed:
                continue
            nodes = None
            # If node is going to be merged with other nodes,
            # collect the set of those nodes' children. This is
            # done to ensure that equivalent children of merged nodes
            # also get merged.
            if node in merges:
                new_node = merges[node]
                nodes = []
                for node_to_merge in inverted_merges[new_node]:
                    nodes.extend(node_to_merge.children)
                processed.extend(inverted_merges[new_node])
            # If node is not going to be merged, simply get the list of
            # node's children.
            else:
                nodes = node.children
                processed.append(node)
            _find_child_merges(nodes)

        return merges

    def merge_nodes(self, merges):
        """Merge some nodes in a graph into others.

        ``merges`` is a dictionary keyed by old nodes, with values equal
        to the nodes that they need to be merged into.  Old nodes'
        parents and children are connected to the new node.

        Arguments:
            merges (dict): dictionary from source nodes -> targets

        """

        def transform(node_list):
            return sorted(set(merges.get(n, n) for n in node_list))

        for old, new in merges.items():
            new.parents = transform(new.parents + old.parents)
            for parent in new.parents:
                parent.children = transform(parent.children)
            new.children = transform(new.children + old.children)
            for child in new.children:
                child.parents = transform(child.parents)
        self.roots = transform(self.roots)

    def normalize(self):
        merges = self.find_merges()
        self.merge_nodes(merges)
        return merges

    def copy(self, old_to_new=None):
        """Create and return a copy of this graph.

        Arguments:
            old_to_new (dict, optional): if provided, this dictionary will
                be populated with mappings from old node -> new node
        """
        # create a mapping dict if one wasn't passed in.
        if old_to_new is None:
            old_to_new = {}

        # first pass creates new nodes
        for node in self.traverse():
            old_to_new[node] = node.copy()

        # second pass hooks up parents and children
        for old, new in old_to_new.items():
            for old_parent in old.parents:
                new.parents.append(old_to_new[old_parent])
            for old_child in old.children:
                new.children.append(old_to_new[old_child])

        graph = Graph([old_to_new[r] for r in self.roots])
        graph.enumerate_traverse()

        return graph

    def union(self, other, old_to_new=None):
        """Create the union of self and other and return it as a new Graph.

        This creates a new graph and does not modify self or other. The
        new Graph has entirely new nodes.

        Arguments:
            other (Graph): another Graph
            old_to_new (dict, optional): if provided, this dictionary will
                be populated with mappings from old node -> new node

        Return:
            (Graph): new Graph containing all nodes and edges from self and other
        """
        if old_to_new is None:
            old_to_new = {}  # mapping from old nodes to new nodes

        def _merge(self_children, other_children, parent):
            """Recursively merge children of self and other.

            Arguments:
                self_children (list or tuple): List of children nodes from self
                other_children (list or tuple): List of children nodes from other
                parent (Node): Parent node for self and other child(ren)

            Modifies old_to_new (dict): Updated dict mapping old nodes from self and other to new
                unioned nodes

            Return:
                (list): list of merged children
            """

            def make_node(*nodes):
                """Make a new node to represent the union of other nodes."""
                new_node = nodes[0].copy()
                for node in nodes:
                    old_to_new[id(node)] = new_node
                return new_node

            new_children = []

            def connect(parent, new_node):
                if parent:
                    parent.add_child(new_node)
                    new_node.add_parent(parent)
                new_children.append(new_node)

            # step through both lists and merge nodes
            self_children, other_children = iter(self_children), iter(other_children)
            self_child = next(self_children, None)
            other_child = next(other_children, None)

            while self_child and other_child:
                if self_child.frame < other_child.frame:
                    # self_child is unique
                    new_node = old_to_new.get(id(self_child))
                    if not new_node:
                        new_node = make_node(self_child)
                        _merge(
                            sorted(self_child.children, key=lambda n: n.frame),
                            (),
                            new_node,
                        )
                    connect(parent, new_node)
                    self_child = next(self_children, None)

                elif self_child.frame > other_child.frame:
                    # other_child is unique
                    new_node = old_to_new.get(id(other_child))
                    if not new_node:
                        new_node = make_node(other_child)
                        _merge(
                            (),
                            sorted(other_child.children, key=lambda n: n.frame),
                            new_node,
                        )
                    connect(parent, new_node)
                    other_child = next(other_children, None)

                else:
                    # self_child and other_child are equal
                    self_mapped = old_to_new.get(id(self_child))
                    other_mapped = old_to_new.get(id(other_child))
                    if not self_mapped and not other_mapped:
                        new_node = make_node(self_child, other_child)
                    else:
                        new_node = self_mapped or other_mapped

                    # map whichever node was not mapped yet
                    if not self_mapped:
                        old_to_new[id(self_child)] = new_node
                        self_side = self_child.children
                    else:
                        self_side = []

                    if not other_mapped:
                        old_to_new[id(other_child)] = new_node
                        other_side = other_child.children
                    else:
                        other_side = []

                    _merge(
                        sorted(self_side, key=lambda n: n.frame),
                        sorted(other_side, key=lambda n: n.frame),
                        new_node,
                    )

                    connect(parent, new_node)
                    self_child = next(self_children, None)
                    other_child = next(other_children, None)

            # finish off whichever list of children is longer
            while self_child:
                new_node = old_to_new.get(id(self_child))
                if not new_node:
                    new_node = make_node(self_child)
                    _merge(
                        sorted(self_child.children, key=lambda n: n.frame),
                        (),
                        new_node,
                    )
                connect(parent, new_node)
                self_child = next(self_children, None)

            while other_child:
                new_node = old_to_new.get(id(other_child))
                if not new_node:
                    new_node = make_node(other_child)
                    _merge(
                        (),
                        sorted(other_child.children, key=lambda n: n.frame),
                        new_node,
                    )
                connect(parent, new_node)
                other_child = next(other_children, None)

            return new_children

        # First establish which nodes correspond to each other
        new_roots = _merge(
            sorted(self.roots, key=lambda n: n.frame),
            sorted(other.roots, key=lambda n: n.frame),
            None,
        )

        graph = Graph(new_roots)
        graph.enumerate_traverse()

        return graph

    def enumerate_depth(self):
        def _iter_depth(node, visited):
            for child in node.children:
                if child not in visited:
                    visited.add(child)
                    # depth of child is depth of node + 1
                    child._depth = node._depth + 1
                    _iter_depth(child, visited)

        visited = set()
        for root in self.roots:
            root._depth = 0  # depth of root node is 0
            _iter_depth(root, visited)

    def enumerate_traverse(self):
        if not self._check_enumerate_traverse():
            for i, node in enumerate(self.traverse()):
                node._hatchet_nid = i

            self.enumerate_depth()

    def _check_enumerate_traverse(self):
        for i, node in enumerate(self.traverse()):
            if i != node._hatchet_nid:
                return False

    def __len__(self):
        """Size of the graph in terms of number of nodes."""
        return sum(1 for _ in self.traverse())

    def __eq__(self, other):
        """Check if two graphs have the same structure by comparing frame at each
        node.
        """
        vs = set()
        vo = set()

        # if both graphs are pointing to the same object, then graphs are equal
        if self is other:
            return True

        # if number of roots do not match, then graphs are not equal
        if len(self.roots) != len(other.roots):
            return False

        if len(self) != len(other):
            return False

        # sort roots by its frame
        ssorted = sorted(self.roots, key=lambda x: x.frame)
        osorted = sorted(other.roots, key=lambda x: x.frame)

        for self_root, other_root in zip(ssorted, osorted):
            # if frames do not match, then nodes are not equal
            if self_root.frame != other_root.frame:
                return False

            if not self_root.dag_equal(other_root, vs, vo):
                return False

        return True

    def __ne__(self, other):
        return not (self == other)

    @staticmethod
    def from_lists(*roots):
        """Convenience method to invoke Node.from_lists() on each root value."""
        if not all(isinstance(r, (list, tuple)) for r in roots):
            raise ValueError(
                "All arguments to Graph.from_lists() must be lists: %s" % roots
            )

        graph = Graph([Node.from_lists(r) for r in roots])
        graph.enumerate_traverse()

        return graph



# Copyright 2017-2023 Lawrence Livermore National Security, LLC and other
# Hatchet Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

import pandas as pd

import hatchet.graphframe
from hatchet.node import Node
from hatchet.graph import Graph
from hatchet.frame import Frame
from hatchet.util.timer import Timer


def _find_child_node(node, name):
    """Return child with given name from parent node"""
    for c in node.children:
        if c.frame.get("name") == name:
            return c
    return None


class SpotDatasetReader:
    """Reads a (single-run) dataset from SpotDB"""

    def __init__(self, regionprofile, metadata, attr_info):
        """Initialize SpotDataset reader

        Args:
            regionprofile (dict): Dict with region names to key:value record with
                metrics. Region names are hierarchical, separated with '/'.
                Example:
                { "a/b/c": { "metric": val, ... }, ... }

            metadata: (dict): Key-value run metadata for this dataset. Example:
                { "launchdate": 123456789, "figure_of_merit": 42.0 }

            attr_info (dict): Information about metric attributes. Contains, e.g.,
                type and alias info. This data is optional. Example:
                { "metric": { "type": "double", "alias": "The Metric", ... }, ... }
        """

        self.regionprofile = regionprofile
        self.attr_info = attr_info
        self.metadata = metadata
        self.df_data = []
        self.roots = {}
        self.metric_columns = set()

        self.timer = Timer()

    def create_graph(self):
        """Create the graph. Fills in df_data and metric_columns."""

        self.df_data.clear()

        for pathstr, vals in self.regionprofile.items():
            # parse { "a/b/c": { "metric": val, ... }, ... } records
            if len(pathstr) == 0:
                continue

            path = pathstr.split("/")
            name = path[-1]
            node = self._create_node(path)

            metrics = {}
            for k, v in vals.items():
                info = self.attr_info.get(k, dict())
                colm = info.get("alias", k)
                type = info.get("type", "string")
                if "inclusive" in k:
                    colm += " (inc)"

                if type == "double":
                    metrics[colm] = float(v)
                elif type == "int" or type == "uint":
                    metrics[colm] = int(v)
                else:
                    metrics[colm] = v
                self.metric_columns.add(colm)

            self.df_data.append(dict({"name": name, "node": node}, **metrics))

    def read(self, default_metric="Total time (inc)"):
        """Create GraphFrame for the given Spot dataset."""

        with self.timer.phase("graph construction"):
            self.create_graph()

        graph = Graph(list(self.roots.values()))
        graph.enumerate_traverse()

        dataframe = pd.DataFrame(data=self.df_data)
        dataframe.set_index(["node"], inplace=True)

        exc_metrics = []
        inc_metrics = []
        for m in self.metric_columns:
            if "(inc)" in m:
                inc_metrics.append(m)
            else:
                exc_metrics.append(m)

        if default_metric not in dataframe.columns:
            if len(inc_metrics) > 0:
                default_metric = inc_metrics[0]
            elif len(exc_metrics) > 0:
                default_metric = exc_metrics[0]

        return hatchet.graphframe.GraphFrame(
            graph,
            dataframe,
            exc_metrics,
            inc_metrics,
            metadata=self.metadata,
            default_metric=default_metric,
        )

    def _create_node(self, path):
        parent = self.roots.get(path[0], None)
        if parent is None:
            parent = Node(Frame(name=path[0]))
            self.roots[path[0]] = parent

        node = parent
        for name in path[1:]:
            node = _find_child_node(parent, name)
            if node is None:
                node = Node(Frame(name=name), parent)
                parent.add_child(node)
            parent = node

        return node


class SpotDBReader:
    """Import multiple runs as graph frames from a SpotDB instance"""

    def __init__(self, db_key, list_of_ids=None, default_metric="Total time (inc)"):
        """Initialize SpotDBReader

        Args:
            db_key (str or SpotDB object): locator for SpotDB instance
                This can be a SpotDB object directly, or a locator for a spot
                database, which is a string with either
                    * A directory for .cali files,
                    * A .sqlite file name
                    * A SQL database URL (e.g., "mysql://hostname/db")

            list_of_ids: The list of run IDs to read from the database.
                If this is None, returns all runs.

            default_metric: Name of the default metric for the GraphFrames.
        """
        self.db_key = db_key
        self.list_of_ids = list_of_ids
        self.default_metric = default_metric

    def read(self):
        """Read given runs from SpotDB

        Returns:
            List of GraphFrames, one for each entry that was found
        """
        import spotdb

        if isinstance(self.db_key, str):
            db = spotdb.connect(self.db_key)
        else:
            db = self.db_key

        runs = self.list_of_ids or db.get_all_run_ids()

        regionprofiles = db.get_regionprofiles(runs)
        metadata = db.get_global_data(runs)
        attr_info = db.get_metric_attribute_metadata()

        result = []

        for run in runs:
            if run in regionprofiles:
                result.append(
                    SpotDatasetReader(
                        regionprofiles[run], metadata[run], attr_info
                    ).read(self.default_metric)
                )

        return result



# Copyright 2017-2023 Lawrence Livermore National Security, LLC and other
# Hatchet Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

from functools import total_ordering


@total_ordering
class Frame:
    """The frame index for a node. The node only stores its frame.

    Arguments:
       attrs (dict): dictionary of attributes and values
    """

    def __init__(self, attrs=None, **kwargs):
        """Construct a frame from a dictionary, or from immediate kwargs.

        Arguments:
            attrs (dict, optional): dictionary of attributes for this Frame

        Keyword arguments are optional, but if they are provided, they
        will be used to update the dictionary.  Keys in kwargs take
        precedence over anything in the attrs dictionary.

        So, these are all functionally equivalent::

            Frame({"name": "foo", "file": "bar.c"})
            Frame(name="foo", file="bar.c")
            Frame({"name": "foo"}, file="bar.c")
            Frame({"name": "foo", "file": "baz.h"}, file="bar.c")

        """
        # attributes dictionary
        self.attrs = attrs if attrs else {}

        # add keyword arguments, if any.
        if kwargs:
            self.attrs.update(kwargs)

        if not self.attrs:
            raise ValueError("Frame must be constructed with attributes!")

        # add type to frame if type is not in the attributes dict or kwargs
        if "type" not in self.attrs:
            self.attrs["type"] = "None"

        self._tuple_repr = None

    def __eq__(self, other):
        return self.tuple_repr == other.tuple_repr

    def __lt__(self, other):
        return self.tuple_repr < other.tuple_repr

    def __gt__(self, other):
        return self.tuple_repr > other.tuple_repr

    def __hash__(self):
        return hash(self.tuple_repr)

    def __str__(self):
        """str() with sorted attributes, so output is deterministic."""
        return "{%s}" % ", ".join("'%s': '%s'" % (k, v) for k, v in self.tuple_repr)

    def __repr__(self):
        return "Frame(%s)" % self

    @property
    def tuple_repr(self):
        """Make a tuple of attributes and values based on reader."""
        if not self._tuple_repr:
            self._tuple_repr = tuple(sorted((k, v) for k, v in self.attrs.items()))
        return self._tuple_repr

    def copy(self):
        return Frame(self.attrs.copy())

    def __getitem__(self, name):
        return self.attrs[name]

    def get(self, name, default=None):
        return self.attrs.get(name, default)

    def values(self, names):
        """Return a tuple of attribute values from this Frame."""
        if isinstance(names, (list, tuple)):
            return tuple(self.attrs.get(name) for name in names)
        else:
            return self.attrs.get(names)



# Copyright 2017-2023 Lawrence Livermore National Security, LLC and other
# Hatchet Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

import pstats
import sys
import pandas as pd


import hatchet.graphframe
from hatchet.node import Node
from hatchet.graph import Graph
from hatchet.frame import Frame


def print_incomptable_msg(stats_file):
    """
    Function which makes the syntax cleaner in Profiler.write_to_file().
    """
    errmsg = """\n Error: Incompatible pstats file ({})\n Please run your code in Python {} to read in this file. \n"""
    if sys.version_info[0] == 2:
        print(errmsg.format(stats_file, 3))
    if sys.version_info[0] == 3:
        print(errmsg.format(stats_file, 2.7))


class StatData:
    """Faux Enum for python"""

    NUMCALLS = 0
    NATIVECALLS = 1
    EXCTIME = 2
    INCTIME = 3
    SRCNODE = 4


class NameData:
    """Faux Enum for python"""

    FILE = 0
    LINE = 1
    FNCNAME = 2


class CProfileReader:
    def __init__(self, filename):
        self.pstats_file = filename

        self.name_to_hnode = {}
        self.name_to_dict = {}

    def _create_node_and_row(self, fn_data, fn_name, stats_dict):
        """
        Description: Takes a profiled function as specified in a pstats file
        and creates a node for it and adds a new line of metadata to our
        dataframe if it does not exist.
        """
        u_fn_name = "{}:{}:{}".format(
            fn_name,
            fn_data[NameData.FILE].split("/")[-1],
            fn_data[NameData.LINE],
        )
        fn_hnode = self.name_to_hnode.get(u_fn_name)

        if not fn_hnode:
            # create a node if it doesn't exist yet
            fn_hnode = Node(Frame({"type": "function", "name": fn_name}), None)
            self.name_to_hnode[u_fn_name] = fn_hnode

            # lookup stat data for source here
            fn_stats = stats_dict[fn_data]
            self._add_node_metadata(u_fn_name, fn_data, fn_stats, fn_hnode)

        return fn_hnode

    def _get_src(self, stat):
        """Gets the source/parent of our current desitnation node"""
        return stat[StatData.SRCNODE]

    def _add_node_metadata(self, stat_name, stat_module_data, stats, hnode):
        """Puts all the metadata associated with a node in a dictionary to insert into pandas."""
        node_dict = {
            "file": stat_module_data[NameData.FILE],
            "line": stat_module_data[NameData.LINE],
            "name": stat_module_data[NameData.FNCNAME],
            "numcalls": stats[StatData.NUMCALLS],
            "nativecalls": stats[StatData.NATIVECALLS],
            "time (inc)": stats[StatData.INCTIME],
            "time": stats[StatData.EXCTIME],
            "node": hnode,
        }
        self.name_to_dict[stat_name] = node_dict

    def create_graph(self):
        """Performs the creation of our node graph"""
        try:
            stats_dict = pstats.Stats(self.pstats_file).__dict__["stats"]
        except ValueError:
            print_incomptable_msg(self.pstats_file)
            raise
        list_roots = []

        # We iterate through each function/node in our stats dict
        for dst_module_data, dst_stats in stats_dict.items():
            dst_name = dst_module_data[NameData.FNCNAME]
            dst_hnode = self._create_node_and_row(dst_module_data, dst_name, stats_dict)

            # get all parents of our current destination node
            # create source nodes and link with destination node
            srcs = self._get_src(dst_stats)
            if srcs == {}:
                list_roots.append(dst_hnode)
            else:
                for src_module_data in srcs.keys():
                    src_name = src_module_data[NameData.FNCNAME]

                    if src_name is not None:
                        src_hnode = self._create_node_and_row(
                            src_module_data, src_name, stats_dict
                        )
                        dst_hnode.add_parent(src_hnode)
                        src_hnode.add_child(dst_hnode)

        return list_roots

    def read(self):
        roots = self.create_graph()
        graph = Graph(roots)
        graph.enumerate_traverse()

        dataframe = pd.DataFrame.from_dict(data=list(self.name_to_dict.values()))
        index = ["node"]
        dataframe.set_index(index, inplace=True)
        dataframe.sort_index(inplace=True)

        return hatchet.graphframe.GraphFrame(graph, dataframe, ["time"], ["time (inc)"])



# Copyright 2017-2023 Lawrence Livermore National Security, LLC and other
# Hatchet Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

from collections import OrderedDict
from contextlib import contextmanager
from datetime import datetime
from io import StringIO


class Timer(object):
    """Simple phase timer with a context manager."""

    def __init__(self):
        self._phase = None
        self._start_time = None
        self._times = OrderedDict()

    def start_phase(self, phase):
        now = datetime.now()
        delta = None

        if self._phase:
            delta = now - self._start_time
            self._times[self._phase] = delta

        self._phase = phase
        self._start_time = now
        return delta

    def end_phase(self):
        assert self._phase and self._start_time

        now = datetime.now()
        delta = now - self._start_time
        if self._times.get(self._phase):
            self._times[self._phase] = self._times.get(self._phase) + delta
        else:
            self._times[self._phase] = delta

        self._phase = None
        self._start_time = None

    def __str__(self):
        out = StringIO()
        out.write("Times:\n")
        for phase, delta in self._times.items():
            out.write("    %-20s %.2fs\n" % (phase + ":", delta.total_seconds()))
        return out.getvalue()

    @contextmanager
    def phase(self, name):
        self.start_phase(name)
        yield
        self.end_phase()



# Copyright 2017-2023 Lawrence Livermore National Security, LLC and other
# Hatchet Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

# This code comes from the equivalent file in Thicket:
# https://github.com/LLNL/thicket
import os
import sys
import subprocess

_filenotfound_errmsg = """
Cannot find NPM!
This is required to use hatchet.vis!
Please install NPM and try again to import hatchet.vis!
"""


def check_npm():
    if sys.version_info[0] == 2:
        from distutils.spawn import find_executable

        return find_executable("npm") is not None
    from shutil import which

    return which("npm") is not None


def npm_build(vis_directory):
    print("Building hatchet.vis using NPM!")
    subprocess.check_output(["npm", "install", "-y"], cwd=vis_directory)
    subprocess.check_output(["npm", "run", "build"], cwd=vis_directory)


# Get the absolute path to this __init__.py file
# Store in curr_dir to minimize the number of variables
curr_dir = os.path.realpath(os.path.expanduser(__file__))
# Get the hatchet/vis directory from the path to this __init__.py file
curr_dir = os.path.abspath(os.path.join(curr_dir, os.pardir))
# Get the path to hatchet/vis/package-lock.json
pkg_lock_file = os.path.abspath(os.path.join(curr_dir, "package-lock.json"))

if not os.path.isfile(pkg_lock_file):
    if not check_npm():
        raise FileNotFoundError(_filenotfound_errmsg)
    else:
        npm_build(curr_dir)



# Copyright 2017-2023 Lawrence Livermore National Security, LLC and other
# Hatchet Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

from hatchet import GraphFrame


def test_graph_equal(mock_graph_literal):
    gf = GraphFrame.from_literal(mock_graph_literal)
    other = GraphFrame.from_literal(mock_graph_literal)

    assert gf.graph == other.graph


def test_graph_not_equal(mock_graph_literal, calc_pi_hpct_db):
    gf = GraphFrame.from_literal(mock_graph_literal)
    other = GraphFrame.from_hpctoolkit(str(calc_pi_hpct_db))

    assert gf.graph != other.graph


def test_dag_not_equal(mock_dag_literal1, mock_dag_literal2):
    gf = GraphFrame.from_literal(mock_dag_literal1)
    other = GraphFrame.from_literal(mock_dag_literal2)

    assert gf.graph != other.graph


def test_union_dag_same_structure(mock_dag_literal1):
    # make graphs g1 and g2 that you know are equal
    gf = GraphFrame.from_literal(mock_dag_literal1)
    other = GraphFrame.from_literal(mock_dag_literal1)

    g1 = gf.graph
    g2 = other.graph

    assert g1 == g2

    g3 = g1.union(g2)
    assert g3 is not g1
    assert g3 is not g2
    assert g3 == g1
    assert g3 == g2



# Copyright 2017-2023 Lawrence Livermore National Security, LLC and other
# Hatchet Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT
try:
    from .roundtrip.roundtrip.manager import Roundtrip

    Roundtrip
except ImportError:
    pass



# Copyright 2017-2023 Lawrence Livermore National Security, LLC and other
# Hatchet Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

import pytest

from hatchet.node import Node, MultiplePathError
from hatchet.frame import Frame
from hatchet.graph import Graph


def test_from_lists():
    node = Node.from_lists("a")
    assert node.frame == Frame(name="a")

    a = Frame(name="a")
    b = Frame(name="b")
    c = Frame(name="c")

    node = Node.from_lists(["a", ["b", "c"]])

    assert node.frame == a
    assert node.children[0].frame == b
    assert node.children[0].children[0].frame == c


def test_from_lists_value_error():
    with pytest.raises(ValueError):
        Node.from_lists(object())


def test_traverse_pre():
    node = Node(Frame(name="a"))
    assert list(node.traverse(attrs="name")) == ["a"]

    node = Node.from_lists(["a", ["b", "d", "e"], ["c", "f", "g"]])
    assert list(node.traverse(attrs="name")) == ["a", "b", "d", "e", "c", "f", "g"]


def test_traverse_post():
    node = Node.from_lists(["a", ["b", "d", "e"], ["c", "f", "g"]])
    assert list(node.traverse(order="post", attrs="name")) == [
        "d",
        "e",
        "b",
        "f",
        "g",
        "c",
        "a",
    ]


def test_traverse_dag():
    d = Node(Frame(name="d"))
    node = Node.from_lists(["a", ["b", d], ["c", d]])
    assert list(node.traverse(attrs="name")) == ["a", "b", "d", "c"]


def test_node_repr():
    d = Node(Frame(a=1, b=2, c=3))
    assert repr(d) == "Node({'a': 1, 'b': 2, 'c': 3, 'type': 'None'})"


def test_path():
    d = Node(Frame(name="d", type="function"))
    node = Node.from_lists(["a", ["b", d]])

    assert d.path() == (
        Node(Frame(name="a")),
        Node(Frame(name="b")),
        Node(Frame(name="d", type="function")),
    )
    assert d.parents[0].path() == (Node(Frame(name="a")), Node(Frame(name="b")))
    assert node.path() == (Node(Frame(name="a")),)


def test_paths():
    d = Node(Frame(name="d"))
    Node.from_lists(["a", ["b", d], ["c", d]])
    with pytest.raises(MultiplePathError):
        d.path()

    assert d.paths() == [
        (Node(Frame(name="a")), Node(Frame(name="b")), Node(Frame(name="d"))),
        (Node(Frame(name="a")), Node(Frame(name="c")), Node(Frame(name="d"))),
    ]


def test_traverse_paths():
    d = Node(Frame(name="d"))
    diamond_subdag = Node.from_lists(("a", ("b", d), ("c", d)))

    g = Graph.from_lists(("e", "f", diamond_subdag), ("g", diamond_subdag, "h"))
    assert list(g.traverse(attrs="name")) == ["e", "a", "b", "d", "c", "f", "g", "h"]


def check_dag_equal():
    chain = Node.from_lists(("a", ("b", ("c", ("d",)))))

    d = Node(Frame(name="d"))
    diamond = Node.from_lists(("a", ("b", d), ("c", d)))

    tree = Node.from_lists(
        ("a", ("b", "e", "f", "g"), ("c", "e", "f", "g"), ("d", "e", "f", "g"))
    )

    assert chain.dag_equal(chain)
    assert chain.dag_equal(chain.copy())

    assert diamond.dag_equal(diamond)
    assert diamond.dag_equal(diamond.copy())

    assert tree.dag_equal(tree)
    assert tree.dag_equal(tree.copy())

    assert not chain.dag_equal(tree)
    assert not chain.dag_equal(diamond)

    assert not tree.dag_equal(chain)
    assert not tree.dag_equal(diamond)

    assert not diamond.dag_equal(chain)
    assert not diamond.dag_equal(tree)



# Copyright 2017-2023 Lawrence Livermore National Security, LLC and other
# Hatchet Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

import json

import numpy as np

from hatchet import GraphFrame
from hatchet.external.console import ConsoleRenderer


def test_graphframe(hatchet_pyinstrument_json):
    """Sanity test a GraphFrame object with known data."""

    def test_children(child_dict, gf):

        df = gf.dataframe.loc[
            (gf.dataframe["name"] == child_dict["function"])
            & (gf.dataframe["file"] == child_dict["file_path_short"])
            & (gf.dataframe["line"] == child_dict["line_no"])
            & (gf.dataframe["time (inc)"] == child_dict["time"])
            & (gf.dataframe["is_application_code"] == child_dict["is_application_code"])
        ]

        assert len(df.index) == 1
        assert df.index[0].frame["name"] == child_dict["function"]

        child_number = 0
        if "children" in child_dict:
            for child in child_dict["children"]:
                # to calculate the number of children
                child_number += 1
                test_children(child, gf)

        # number of children should be the same.
        assert len(df.index[0].children) == child_number

    # create a graphframe using from_pyinstrument
    gf = GraphFrame.from_pyinstrument(str(hatchet_pyinstrument_json))

    graph_dict = []
    # read directly from the input file to compare it with the graphframe.
    with open(str(hatchet_pyinstrument_json)) as pyinstrument_json:
        graph_dict = json.load(pyinstrument_json)

        # roots should be the same
        assert graph_dict["root_frame"]["function"] == gf.graph.roots[0].frame["name"]

        child_number = 0
        if "children" in graph_dict["root_frame"]:
            for child in graph_dict["root_frame"]["children"]:
                # to calculate the number of children
                child_number += 1
                test_children(child, gf)

        # number of children should be the same.
        assert len(gf.graph.roots[0].children) == child_number

    assert len(gf.dataframe.groupby("name")) == 44

    gf.dataframe.reset_index(inplace=True)

    for col in gf.dataframe.columns:
        if col in ("time (inc)", "time"):
            assert gf.dataframe[col].dtype == np.float64
        elif col in ("line"):
            assert gf.dataframe[col].dtype == np.int64
        elif col in ("is_application_code"):
            assert gf.dataframe[col].dtype == bool
        elif col in ("name", "type", "file", "node"):
            assert gf.dataframe[col].dtype == object


def test_tree(hatchet_pyinstrument_json):
    """Sanity test a GraphFrame object with known data."""
    gf = GraphFrame.from_pyinstrument(str(hatchet_pyinstrument_json))

    output = ConsoleRenderer(unicode=True, color=False).render(
        gf.graph.roots,
        gf.dataframe,
        metric_column="time",
        precision=3,
        name_column="name",
        expand_name=False,
        context_column="file",
        rank=0,
        thread=0,
        depth=10000,
        highlight_name=False,
        colormap="RdYlGn",
        invert_colormap=False,
        render_header=True,
    )
    assert "0.000 <module> examples.py" in output
    assert "0.025 read hatchet/readers/caliper_reader.py" in output

    output = ConsoleRenderer(unicode=True, color=False).render(
        gf.graph.roots,
        gf.dataframe,
        metric_column="time (inc)",
        precision=3,
        name_column="name",
        expand_name=False,
        context_column="file",
        rank=0,
        thread=0,
        depth=10000,
        highlight_name=False,
        colormap="RdYlGn",
        invert_colormap=False,
        render_header=True,
    )
    assert "0.478 <module> examples.py" in output
    assert "0.063 from_caliper_json hatchet/graphframe.py" in output


def test_graphframe_to_literal(hatchet_pyinstrument_json):
    """Sanity test a GraphFrame object with known data."""
    gf = GraphFrame.from_pyinstrument(str(hatchet_pyinstrument_json))
    graph_literal = gf.to_literal()

    gf2 = GraphFrame.from_literal(graph_literal)

    assert len(gf.graph) == len(gf2.graph)



# Copyright 2017-2023 Lawrence Livermore National Security, LLC and other
# Hatchet Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

from numbers import Real
import re
import sys
import pandas as pd  # noqa: F401
from pandas.api.types import is_numeric_dtype, is_string_dtype  # noqa: F401
import numpy as np  # noqa: F401
from textx import metamodel_from_str
from textx.exceptions import TextXError
import warnings

from .errors import InvalidQueryPath, InvalidQueryFilter, RedundantQueryFilterWarning
from .query import Query


# PEG grammar for the String-based dialect
CYPHER_GRAMMAR = u"""
FullQuery: path_expr=MatchExpr(cond_expr=WhereExpr)?;
MatchExpr: 'MATCH' path=PathQuery;
PathQuery: '(' nodes=NodeExpr ')'('->' '(' nodes=NodeExpr ')')*;
NodeExpr: ((wcard=INT | wcard=STRING) ',' name=ID) | (wcard=INT | wcard=STRING) |  name=ID;
WhereExpr: 'WHERE' ConditionExpr;
ConditionExpr: conditions+=CompoundCond;
CompoundCond: UnaryCond | BinaryCond;
BinaryCond: AndCond | OrCond;
AndCond: 'AND' subcond=UnaryCond;
OrCond: 'OR' subcond=UnaryCond;
UnaryCond: NotCond | SingleCond;
NotCond: 'NOT' subcond=SingleCond;
SingleCond: StringCond | NumberCond | NoneCond | NotNoneCond | LeafCond | NotLeafCond;
NoneCond: name=ID '.' prop=STRING 'IS NONE';
NotNoneCond: name=ID '.' prop=STRING 'IS NOT NONE';
LeafCond: name=ID 'IS LEAF';
NotLeafCond: name=ID 'IS NOT LEAF';
StringCond: StringEq | StringStartsWith | StringEndsWith | StringContains | StringMatch;
StringEq: name=ID '.' prop=STRING '=' val=STRING;
StringStartsWith: name=ID '.' prop=STRING 'STARTS WITH' val=STRING;
StringEndsWith: name=ID '.' prop=STRING 'ENDS WITH' val=STRING;
StringContains: name=ID '.' prop=STRING 'CONTAINS' val=STRING;
StringMatch: name=ID '.' prop=STRING '=~' val=STRING;
NumberCond: NumEq | NumLt | NumGt | NumLte | NumGte | NumNan | NumNotNan | NumInf | NumNotInf;
NumEq: name=ID '.' prop=STRING '=' val=NUMBER;
NumLt: name=ID '.' prop=STRING '<' val=NUMBER;
NumGt: name=ID '.' prop=STRING '>' val=NUMBER;
NumLte: name=ID '.' prop=STRING '<=' val=NUMBER;
NumGte: name=ID '.' prop=STRING '>=' val=NUMBER;
NumNan: name=ID '.' prop=STRING 'IS NAN';
NumNotNan: name=ID '.' prop=STRING 'IS NOT NAN';
NumInf: name=ID '.' prop=STRING 'IS INF';
NumNotInf: name=ID '.' prop=STRING 'IS NOT INF';
"""

# TextX metamodel for the String-based dialect
cypher_query_mm = metamodel_from_str(CYPHER_GRAMMAR)


def cname(obj):
    """Utility function to get the name of the rule represented by the input"""
    return obj.__class__.__name__


def filter_check_types(type_check, df_row, filt_lambda):
    """Utility function used in String-based predicates
       to make sure the node data used in the actual boolean predicate
       is of the correct type.

    Arguments:
        type_check (str): a string containing a boolean Python expression used to validate node data typing
        df_row (pandas.Series or pandas.DataFrame): the row (or sub-DataFrame) representing the data for the current node being tested
        filt_lambda (Callable): the lambda used to actually confirm whether the node satisfies the predicate

    Returns:
        (bool): True if the node satisfies the predicate. False otherwise
    """
    try:
        if type_check == "" or eval(type_check):
            return filt_lambda(df_row)
        else:
            raise InvalidQueryFilter("Type mismatch in filter")
    except KeyError:
        return False


class StringQuery(Query):

    """Class for representing and parsing queries using the String-based dialect."""

    def __init__(self, cypher_query, multi_index_mode="off"):
        """Builds a new StringQuery object representing a query in the String-based dialect.

        Arguments:
            cypher_query (str): a query in the String-based dialect
        """
        if sys.version_info[0] == 2:
            super(StringQuery, self).__init__()
        else:
            super().__init__()
        assert multi_index_mode in ["off", "all", "any"]
        self.multi_index_mode = multi_index_mode
        model = None
        try:
            model = cypher_query_mm.model_from_str(cypher_query)
        except TextXError as e:
            # TODO Change to a "raise-from" expression when Python 2.7 support is dropped
            raise InvalidQueryPath(
                "Invalid String Dialect Query Detected. Parser Error Message: {}".format(
                    e.message
                )
            )
        self.wcards = []
        self.wcard_pos = {}
        self._parse_path(model.path_expr)
        self.filters = [[] for _ in self.wcards]
        self._parse_conditions(model.cond_expr)
        self.lambda_filters = [None for _ in self.wcards]
        self._build_lambdas()
        self._build_query()

    def _build_query(self):
        """Builds the entire query using 'match' and 'rel' using
        the pre-parsed quantifiers and predicates.
        """
        for i in range(0, len(self.wcards)):
            wcard = self.wcards[i][0]
            # TODO Remove this when Python 2.7 support is dropped.
            if sys.version_info[0] == 2 and not isinstance(wcard, Real):
                wcard = wcard.encode("ascii", "ignore")
            filt_str = self.lambda_filters[i]
            if filt_str is None:
                if i == 0:
                    self.match(quantifier=wcard)
                else:
                    self.rel(quantifier=wcard)
            else:
                if i == 0:
                    self.match(quantifier=wcard, predicate=eval(filt_str))
                else:
                    self.rel(quantifier=wcard, predicate=eval(filt_str))

    def _build_lambdas(self):
        """Constructs the final predicate lambdas from the pre-parsed
        predicate information.
        """
        for i in range(0, len(self.wcards)):
            n = self.wcards[i]
            if n[1] != "":
                bool_expr = ""
                type_check = ""
                for j, cond in enumerate(self.filters[i]):
                    if cond[0] is not None:
                        bool_expr += " {}".format(cond[0])
                    bool_expr += " {}".format(cond[1])
                    if cond[2] is not None:
                        if j == 0:
                            type_check += " {}".format(cond[2])
                        else:
                            type_check += " and {}".format(cond[2])
                bool_expr = "lambda df_row: {}".format(bool_expr)
                bool_expr = (
                    'lambda df_row: filter_check_types("{}", df_row, {})'.format(
                        type_check, bool_expr
                    )
                )
                self.lambda_filters[i] = bool_expr

    def _parse_path(self, path_obj):
        """Parses the MATCH statement of a String-based query."""
        nodes = path_obj.path.nodes
        idx = len(self.wcards)
        for n in nodes:
            new_node = [n.wcard, n.name]
            if n.wcard is None or n.wcard == "" or n.wcard == 0:
                new_node[0] = "."
            self.wcards.append(new_node)
            if n.name != "":
                self.wcard_pos[n.name] = idx
            idx += 1

    def _parse_conditions(self, cond_expr):
        """Top level function for parsing the WHERE statement of
        a String-based query.
        """
        conditions = cond_expr.conditions
        for cond in conditions:
            converted_condition = None
            if self._is_unary_cond(cond):
                converted_condition = self._parse_unary_cond(cond)
            elif self._is_binary_cond(cond):
                converted_condition = self._parse_binary_cond(cond)
            else:
                raise RuntimeError("Bad Condition")
            self.filters[self.wcard_pos[converted_condition[1]]].append(
                [converted_condition[0], converted_condition[2], converted_condition[3]]
            )
        for i in range(0, len(self.filters)):
            if len(self.filters[i]) > 0:
                if self.filters[i][0][0] != "not":
                    self.filters[i][0][0] = None

    def _is_unary_cond(self, obj):
        """Detect whether a predicate is unary or not."""
        if (
            cname(obj) == "NotCond"
            or self._is_str_cond(obj)
            or self._is_num_cond(obj)
            or cname(obj) in ["NoneCond", "NotNoneCond", "LeafCond", "NotLeafCond"]
        ):
            return True
        return False

    def _is_binary_cond(self, obj):
        """Detect whether a predicate is binary or not."""
        if cname(obj) in ["AndCond", "OrCond"]:
            return True
        return False

    def _parse_binary_cond(self, obj):
        """Top level function for parsing binary predicates."""
        if cname(obj) == "AndCond":
            return self._parse_and_cond(obj)
        if cname(obj) == "OrCond":
            return self._parse_or_cond(obj)
        raise RuntimeError("Bad Binary Condition")

    def _parse_or_cond(self, obj):
        """Top level function for parsing predicates combined with logical OR."""
        converted_subcond = self._parse_unary_cond(obj.subcond)
        converted_subcond[0] = "or"
        return converted_subcond

    def _parse_and_cond(self, obj):
        """Top level function for parsing predicates combined with logical AND."""
        converted_subcond = self._parse_unary_cond(obj.subcond)
        converted_subcond[0] = "and"
        return converted_subcond

    def _parse_unary_cond(self, obj):
        """Top level function for parsing unary predicates."""
        if cname(obj) == "NotCond":
            return self._parse_not_cond(obj)
        return self._parse_single_cond(obj)

    def _parse_not_cond(self, obj):
        """Parse predicates containing the logical NOT operator."""
        converted_subcond = self._parse_single_cond(obj.subcond)
        converted_subcond[2] = "not {}".format(converted_subcond[2])
        return converted_subcond

    def _run_method_based_on_multi_idx_mode(self, method_name, obj):
        real_method_name = method_name
        if self.multi_index_mode != "off":
            real_method_name = method_name + "_multi_idx"
        method = eval("StringQuery.{}".format(real_method_name))
        return method(self, obj)

    def _parse_single_cond(self, obj):
        """Top level function for parsing individual numeric or string predicates."""
        if self._is_str_cond(obj):
            return self._parse_str(obj)
        if self._is_num_cond(obj):
            return self._parse_num(obj)
        if cname(obj) == "NoneCond":
            return self._run_method_based_on_multi_idx_mode("_parse_none", obj)
        if cname(obj) == "NotNoneCond":
            return self._run_method_based_on_multi_idx_mode("_parse_not_none", obj)
        if cname(obj) == "LeafCond":
            return self._run_method_based_on_multi_idx_mode("_parse_leaf", obj)
        if cname(obj) == "NotLeafCond":
            return self._run_method_based_on_multi_idx_mode("_parse_not_leaf", obj)
        raise RuntimeError("Bad Single Condition")

    def _parse_none(self, obj):
        """Parses 'property IS NONE'."""
        if obj.prop == "depth":
            return [
                None,
                obj.name,
                "df_row.name._depth is None",
                None,
            ]
        if obj.prop == "node_id":
            return [
                None,
                obj.name,
                "df_row.name._hatchet_nid is None",
                None,
            ]
        return [
            None,
            obj.name,
            'df_row["{}"] is None'.format(obj.prop),
            None,
        ]

    def _add_aggregation_call_to_multi_idx_predicate(self, predicate):
        if self.multi_index_mode == "any":
            return predicate + ".any()"
        return predicate + ".all()"

    def _parse_none_multi_idx(self, obj):
        if obj.prop == "depth":
            return [
                None,
                obj.name,
                "df_row.index.get_level_values('node')[0]._depth is None",
                None,
            ]
        if obj.prop == "node_id":
            return [
                None,
                obj.name,
                "df_row.index.get_level_values('node')[0]._hatchet_nid is None",
                None,
            ]
        if self.multi_index_mode == "any":
            return [
                None,
                obj.name,
                "df_row['{}'].apply(lambda elem: elem is None).any()".format(obj.prop),
                None,
            ]
        # if self.multi_index_mode == "all":
        return [
            None,
            obj.name,
            self._add_aggregation_call_to_multi_idx_predicate(
                "df_row['{}'].apply(lambda elem: elem is None)".format(obj.prop)
            ),
            None,
        ]

    def _parse_not_none(self, obj):
        """Parses 'property IS NOT NONE'."""
        if obj.prop == "depth":
            return [
                None,
                obj.name,
                "df_row.name._depth is not None",
                None,
            ]
        if obj.prop == "node_id":
            return [
                None,
                obj.name,
                "df_row.name._hatchet_nid is not None",
                None,
            ]
        return [
            None,
            obj.name,
            'df_row["{}"] is not None'.format(obj.prop),
            None,
        ]

    def _parse_not_none_multi_idx(self, obj):
        if obj.prop == "depth":
            return [
                None,
                obj.name,
                "df_row.index.get_level_values('node')[0]._depth is not None",
                None,
            ]
        if obj.prop == "node_id":
            return [
                None,
                obj.name,
                "df_row.index.get_level_values('node')[0]._hatchet_nid is not None",
                None,
            ]
        return [
            None,
            obj.name,
            self._add_aggregation_call_to_multi_idx_predicate(
                "df_row['{}'].apply(lambda elem: elem is not None)".format(obj.prop)
            ),
            None,
        ]

    def _parse_leaf(self, obj):
        """Parses 'node IS LEAF'."""
        return [
            None,
            obj.name,
            "len(df_row.name.children) == 0",
            None,
        ]

    def _parse_leaf_multi_idx(self, obj):
        return [
            None,
            obj.name,
            "len(df_row.index.get_level_values('node')[0].children) == 0",
            None,
        ]

    def _parse_not_leaf(self, obj):
        """Parses 'node IS NOT LEAF'."""
        return [
            None,
            obj.name,
            "len(df_row.name.children) > 0",
            None,
        ]

    def _parse_not_leaf_multi_idx(self, obj):
        return [
            None,
            obj.name,
            "len(df_row.index.get_level_values('node')[0].children) > 0",
            None,
        ]

    def _is_str_cond(self, obj):
        """Determines whether a predicate is for string data."""
        if cname(obj) in [
            "StringEq",
            "StringStartsWith",
            "StringEndsWith",
            "StringContains",
            "StringMatch",
        ]:
            return True
        return False

    def _is_num_cond(self, obj):
        """Determines whether a predicate is for numeric data."""
        if cname(obj) in [
            "NumEq",
            "NumLt",
            "NumGt",
            "NumLte",
            "NumGte",
            "NumNan",
            "NumNotNan",
            "NumInf",
            "NumNotInf",
        ]:
            return True
        return False

    def _parse_str(self, obj):
        """Function that redirects processing of string predicates
        to the correct function.
        """
        if cname(obj) == "StringEq":
            return self._run_method_based_on_multi_idx_mode("_parse_str_eq", obj)
        if cname(obj) == "StringStartsWith":
            return self._run_method_based_on_multi_idx_mode(
                "_parse_str_starts_with", obj
            )
        if cname(obj) == "StringEndsWith":
            return self._run_method_based_on_multi_idx_mode("_parse_str_ends_with", obj)
        if cname(obj) == "StringContains":
            return self._run_method_based_on_multi_idx_mode("_parse_str_contains", obj)
        if cname(obj) == "StringMatch":
            return self._run_method_based_on_multi_idx_mode("_parse_str_match", obj)
        raise RuntimeError("Bad String Op Class")

    def _parse_str_eq(self, obj):
        """Processes string equivalence predicates."""
        return [
            None,
            obj.name,
            'df_row["{}"] == "{}"'.format(obj.prop, obj.val),
            "isinstance(df_row['{}'], str)".format(obj.prop),
        ]

    def _parse_str_eq_multi_idx(self, obj):
        return [
            None,
            obj.name,
            self._add_aggregation_call_to_multi_idx_predicate(
                'df_row["{}"].apply(lambda elem: elem == "{}")'.format(
                    obj.prop, obj.val
                )
            ),
            "is_string_dtype(df_row['{}'])".format(obj.prop),
        ]

    def _parse_str_starts_with(self, obj):
        """Processes string 'startswith' predicates."""
        return [
            None,
            obj.name,
            'df_row["{}"].startswith("{}")'.format(obj.prop, obj.val),
            "isinstance(df_row['{}'], str)".format(obj.prop),
        ]

    def _parse_str_starts_with_multi_idx(self, obj):
        return [
            None,
            obj.name,
            self._add_aggregation_call_to_multi_idx_predicate(
                'df_row["{}"].apply(lambda elem: elem.startswith("{}"))'.format(
                    obj.prop, obj.val
                )
            ),
            "is_string_dtype(df_row['{}'])".format(obj.prop),
        ]

    def _parse_str_ends_with(self, obj):
        """Processes string 'endswith' predicates."""
        return [
            None,
            obj.name,
            'df_row["{}"].endswith("{}")'.format(obj.prop, obj.val),
            "isinstance(df_row['{}'], str)".format(obj.prop),
        ]

    def _parse_str_ends_with_multi_idx(self, obj):
        return [
            None,
            obj.name,
            self._add_aggregation_call_to_multi_idx_predicate(
                'df_row["{}"].apply(lambda elem: elem.endswith("{}"))'.format(
                    obj.prop, obj.val
                )
            ),
            "is_string_dtype(df_row['{}'])".format(obj.prop),
        ]

    def _parse_str_contains(self, obj):
        """Processes string 'contains' predicates."""
        return [
            None,
            obj.name,
            '"{}" in df_row["{}"]'.format(obj.val, obj.prop),
            "isinstance(df_row['{}'], str)".format(obj.prop),
        ]

    def _parse_str_contains_multi_idx(self, obj):
        return [
            None,
            obj.name,
            self._add_aggregation_call_to_multi_idx_predicate(
                'df_row["{}"].apply(lambda elem: "{}" in elem)'.format(
                    obj.prop, obj.val
                )
            ),
            "is_string_dtype(df_row['{}'])".format(obj.prop),
        ]

    def _parse_str_match(self, obj):
        """Processes string regex match predicates."""
        return [
            None,
            obj.name,
            're.match("{}", df_row["{}"]) is not None'.format(obj.val, obj.prop),
            "isinstance(df_row['{}'], str)".format(obj.prop),
        ]

    def _parse_str_match_multi_idx(self, obj):
        return [
            None,
            obj.name,
            self._add_aggregation_call_to_multi_idx_predicate(
                'df_row["{}"].apply(lambda elem: re.match("{}", elem) is not None)'.format(
                    obj.prop, obj.val
                )
            ),
            "is_string_dtype(df_row['{}'])".format(obj.prop),
        ]

    def _parse_num(self, obj):
        """Function that redirects processing of numeric predicates
        to the correct function.
        """
        if cname(obj) == "NumEq":
            return self._run_method_based_on_multi_idx_mode("_parse_num_eq", obj)
        if cname(obj) == "NumLt":
            return self._run_method_based_on_multi_idx_mode("_parse_num_lt", obj)
        if cname(obj) == "NumGt":
            return self._run_method_based_on_multi_idx_mode("_parse_num_gt", obj)
        if cname(obj) == "NumLte":
            return self._run_method_based_on_multi_idx_mode("_parse_num_lte", obj)
        if cname(obj) == "NumGte":
            return self._run_method_based_on_multi_idx_mode("_parse_num_gte", obj)
        if cname(obj) == "NumNan":
            return self._run_method_based_on_multi_idx_mode("_parse_num_nan", obj)
        if cname(obj) == "NumNotNan":
            return self._run_method_based_on_multi_idx_mode("_parse_num_not_nan", obj)
        if cname(obj) == "NumInf":
            return self._run_method_based_on_multi_idx_mode("_parse_num_inf", obj)
        if cname(obj) == "NumNotInf":
            return self._run_method_based_on_multi_idx_mode("_parse_num_not_inf", obj)
        raise RuntimeError("Bad Number Op Class")

    def _parse_num_eq(self, obj):
        """Processes numeric equivalence predicates."""
        if obj.prop == "depth":
            if obj.val == -1:
                return [
                    None,
                    obj.name,
                    "len(df_row.name.children) == 0",
                    None,
                ]
            elif obj.val < 0:
                warnings.warn(
                    """
                    The 'depth' property of a Node is strictly non-negative.
                    This condition will always be false.
                    The statement that triggered this warning is:
                    {}
                    """.format(
                        obj
                    ),
                    RedundantQueryFilterWarning,
                )
                return [
                    None,
                    obj.name,
                    "False",
                    "isinstance(df_row.name._depth, Real)",
                ]
            return [
                None,
                obj.name,
                "df_row.name._depth == {}".format(obj.val),
                "isinstance(df_row.name._depth, Real)",
            ]
        if obj.prop == "node_id":
            if obj.val < 0:
                warnings.warn(
                    """
                    The 'node_id' property of a Node is strictly non-negative.
                    This condition will always be false.
                    The statement that triggered this warning is:
                    {}
                    """.format(
                        obj
                    ),
                    RedundantQueryFilterWarning,
                )
                return [
                    None,
                    obj.name,
                    "False",
                    "isinstance(df_row.name._hatchet_nid, Real)",
                ]
            return [
                None,
                obj.name,
                "df_row.name._hatchet_nid == {}".format(obj.val),
                "isinstance(df_row.name._hatchet_nid, Real)",
            ]
        return [
            None,
            obj.name,
            'df_row["{}"] == {}'.format(obj.prop, obj.val),
            "isinstance(df_row['{}'], Real)".format(obj.prop),
        ]

    def _parse_num_eq_multi_idx(self, obj):
        if obj.prop == "depth":
            if obj.val == -1:
                return [
                    None,
                    obj.name,
                    "len(df_row.index.get_level_values('node')[0].children) == 0",
                    None,
                ]
            elif obj.val < 0:
                warnings.warn(
                    """
                    The 'depth' property of a Node is strictly non-negative.
                    This condition will always be false.
                    The statement that triggered this warning is:
                    {}
                    """.format(
                        obj
                    ),
                    RedundantQueryFilterWarning,
                )
                return [
                    None,
                    obj.name,
                    "False",
                    "isinstance(df_row.index.get_level_values('node')[0]._depth, Real)",
                ]
            return [
                None,
                obj.name,
                "df_row.index.get_level_values('node')[0]._depth == {}".format(obj.val),
                "isinstance(df_row.index.get_level_values('node')[0]._depth, Real)",
            ]
        if obj.prop == "node_id":
            if obj.val < 0:
                warnings.warn(
                    """
                    The 'node_id' property of a Node is strictly non-negative.
                    This condition will always be false.
                    The statement that triggered this warning is:
                    {}
                    """.format(
                        obj
                    ),
                    RedundantQueryFilterWarning,
                )
                return [
                    None,
                    obj.name,
                    "False",
                    "isinstance(df_row.index.get_level_values('node')[0]._hatchet_nid, Real)",
                ]
            return [
                None,
                obj.name,
                "df_row.index.get_level_values('node')[0]._hatchet_nid == {}".format(
                    obj.val
                ),
                "isinstance(df_row.index.get_level_values('node')[0]._hatchet_nid, Real)",
            ]
        return [
            None,
            obj.name,
            self._add_aggregation_call_to_multi_idx_predicate(
                'df_row["{}"].apply(lambda elem: elem == {})'.format(obj.prop, obj.val)
            ),
            "is_numeric_dtype(df_row['{}'])".format(obj.prop),
        ]

    def _parse_num_lt(self, obj):
        """Processes numeric less-than predicates."""
        if obj.prop == "depth":
            if obj.val < 0:
                warnings.warn(
                    """
                    The 'depth' property of a Node is strictly non-negative.
                    This condition will always be false.
                    The statement that triggered this warning is:
                    {}
                    """.format(
                        obj
                    ),
                    RedundantQueryFilterWarning,
                )
                return [
                    None,
                    obj.name,
                    "False",
                    "isinstance(df_row.name._depth, Real)",
                ]
            return [
                None,
                obj.name,
                "df_row.name._depth < {}".format(obj.val),
                "isinstance(df_row.name._depth, Real)",
            ]
        if obj.prop == "node_id":
            if obj.val < 0:
                warnings.warn(
                    """
                    The 'node_id' property of a Node is strictly non-negative.
                    This condition will always be false.
                    The statement that triggered this warning is:
                    {}
                    """.format(
                        obj
                    ),
                    RedundantQueryFilterWarning,
                )
                return [
                    None,
                    obj.name,
                    "False",
                    "isinstance(df_row.name._hatchet_nid, Real)",
                ]
            return [
                None,
                obj.name,
                "df_row.name._hatchet_nid < {}".format(obj.val),
                "isinstance(df_row.name._hatchet_nid, Real)",
            ]
        return [
            None,
            obj.name,
            'df_row["{}"] < {}'.format(obj.prop, obj.val),
            "isinstance(df_row['{}'], Real)".format(obj.prop),
        ]

    def _parse_num_lt_multi_idx(self, obj):
        if obj.prop == "depth":
            if obj.val < 0:
                warnings.warn(
                    """
                    The 'depth' property of a Node is strictly non-negative.
                    This condition will always be false.
                    The statement that triggered this warning is:
                    {}
                    """.format(
                        obj
                    ),
                    RedundantQueryFilterWarning,
                )
                return [
                    None,
                    obj.name,
                    "False",
                    "isinstance(df_row.index.get_level_values('node')[0]._depth, Real)",
                ]
            return [
                None,
                obj.name,
                "df_row.index.get_level_values('node')[0]._depth < {}".format(obj.val),
                "isinstance(df_row.index.get_level_values('node')[0]._depth, Real)",
            ]
        if obj.prop == "node_id":
            if obj.val < 0:
                warnings.warn(
                    """
                    The 'node_id' property of a Node is strictly non-negative.
                    This condition will always be false.
                    The statement that triggered this warning is:
                    {}
                    """.format(
                        obj
                    ),
                    RedundantQueryFilterWarning,
                )
                return [
                    None,
                    obj.name,
                    "False",
                    "isinstance(df_row.index.get_level_values('node')[0]._hatchet_nid, Real)",
                ]
            return [
                None,
                obj.name,
                "df_row.index.get_level_values('node')[0]._hatchet_nid < {}".format(
                    obj.val
                ),
                "isinstance(df_row.index.get_level_values('node')[0]._hatchet_nid, Real)",
            ]
        return [
            None,
            obj.name,
            self._add_aggregation_call_to_multi_idx_predicate(
                'df_row["{}"].apply(lambda elem: elem < {})'.format(obj.prop, obj.val)
            ),
            "is_numeric_dtype(df_row['{}'])".format(obj.prop),
        ]

    def _parse_num_gt(self, obj):
        """Processes numeric greater-than predicates."""
        if obj.prop == "depth":
            if obj.val < 0:
                warnings.warn(
                    """
                    The 'depth' property of a Node is strictly non-negative.
                    This condition will always be true.
                    The statement that triggered this warning is:
                    {}
                    """.format(
                        obj
                    ),
                    RedundantQueryFilterWarning,
                )
                return [
                    None,
                    obj.name,
                    "True",
                    "isinstance(df_row.name._depth, Real)",
                ]
            return [
                None,
                obj.name,
                "df_row.name._depth > {}".format(obj.val),
                "isinstance(df_row.name._depth, Real)",
            ]
        if obj.prop == "node_id":
            if obj.val < 0:
                warnings.warn(
                    """
                    The 'node_id' property of a Node is strictly non-negative.
                    This condition will always be true.
                    The statement that triggered this warning is:
                    {}
                    """.format(
                        obj
                    ),
                    RedundantQueryFilterWarning,
                )
                return [
                    None,
                    obj.name,
                    "True",
                    "isinstance(df_row.name._hatchet_nid, Real)",
                ]
            return [
                None,
                obj.name,
                "df_row.name._hatchet_nid > {}".format(obj.val),
                "isinstance(df_row.name._hatchet_nid, Real)",
            ]
        return [
            None,
            obj.name,
            'df_row["{}"] > {}'.format(obj.prop, obj.val),
            "isinstance(df_row['{}'], Real)".format(obj.prop),
        ]

    def _parse_num_gt_multi_idx(self, obj):
        if obj.prop == "depth":
            if obj.val < 0:
                warnings.warn(
                    """
                    The 'depth' property of a Node is strictly non-negative.
                    This condition will always be true.
                    The statement that triggered this warning is:
                    {}
                    """.format(
                        obj
                    ),
                    RedundantQueryFilterWarning,
                )
                return [
                    None,
                    obj.name,
                    "True",
                    "isinstance(df_row.index.get_level_values('node')[0]._depth, Real)",
                ]
            return [
                None,
                obj.name,
                "df_row.index.get_level_values('node')[0]._depth > {}".format(obj.val),
                "isinstance(df_row.index.get_level_values('node')[0]._depth, Real)",
            ]
        if obj.prop == "node_id":
            if obj.val < 0:
                warnings.warn(
                    """
                    The 'node_id' property of a Node is strictly non-negative.
                    This condition will always be true.
                    The statement that triggered this warning is:
                    {}
                    """.format(
                        obj
                    ),
                    RedundantQueryFilterWarning,
                )
                return [
                    None,
                    obj.name,
                    "True",
                    "isinstance(df_row.index.get_level_values('node')[0]._hatchet_nid, Real)",
                ]
            return [
                None,
                obj.name,
                "df_row.index.get_level_values('node')[0]._hatchet_nid > {}".format(
                    obj.val
                ),
                "isinstance(df_row.index.get_level_values('node')[0]._hatchet_nid, Real)",
            ]
        return [
            None,
            obj.name,
            self._add_aggregation_call_to_multi_idx_predicate(
                'df_row["{}"].apply(lambda elem: elem > {})'.format(obj.prop, obj.val)
            ),
            "is_numeric_dtype(df_row['{}'])".format(obj.prop),
        ]

    def _parse_num_lte(self, obj):
        """Processes numeric less-than-or-equal-to predicates."""
        if obj.prop == "depth":
            if obj.val < 0:
                warnings.warn(
                    """
                    The 'depth' property of a Node is strictly non-negative.
                    This condition will always be false.
                    The statement that triggered this warning is:
                    {}
                    """.format(
                        obj
                    ),
                    RedundantQueryFilterWarning,
                )
                return [
                    None,
                    obj.name,
                    "False",
                    "isinstance(df_row.name._depth, Real)",
                ]
            return [
                None,
                obj.name,
                "df_row.name._depth <= {}".format(obj.val),
                "isinstance(df_row.name._depth, Real)",
            ]
        if obj.prop == "node_id":
            if obj.val < 0:
                warnings.warn(
                    """
                    The 'node_id' property of a Node is strictly non-negative.
                    This condition will always be false.
                    The statement that triggered this warning is:
                    {}
                    """.format(
                        obj
                    ),
                    RedundantQueryFilterWarning,
                )
                return [
                    None,
                    obj.name,
                    "False",
                    "isinstance(df_row.name._hatchet_nid, Real)",
                ]
            return [
                None,
                obj.name,
                "df_row.name._hatchet_nid <= {}".format(obj.val),
                "isinstance(df_row.name._hatchet_nid, Real)",
            ]
        return [
            None,
            obj.name,
            'df_row["{}"] <= {}'.format(obj.prop, obj.val),
            "isinstance(df_row['{}'], Real)".format(obj.prop),
        ]

    def _parse_num_lte_multi_idx(self, obj):
        if obj.prop == "depth":
            if obj.val < 0:
                warnings.warn(
                    """
                    The 'depth' property of a Node is strictly non-negative.
                    This condition will always be false.
                    The statement that triggered this warning is:
                    {}
                    """.format(
                        obj
                    ),
                    RedundantQueryFilterWarning,
                )
                return [
                    None,
                    obj.name,
                    "False",
                    "isinstance(df_row.index.get_level_values('node')[0]._depth, Real)",
                ]
            return [
                None,
                obj.name,
                "df_row.index.get_level_values('node')[0]._depth <= {}".format(obj.val),
                "isinstance(df_row.index.get_level_values('node')[0]._depth, Real)",
            ]
        if obj.prop == "node_id":
            if obj.val < 0:
                warnings.warn(
                    """
                    The 'node_id' property of a Node is strictly non-negative.
                    This condition will always be false.
                    The statement that triggered this warning is:
                    {}
                    """.format(
                        obj
                    ),
                    RedundantQueryFilterWarning,
                )
                return [
                    None,
                    obj.name,
                    "False",
                    "isinstance(df_row.index.get_level_values('node')[0]._hatchet_nid, Real)",
                ]
            return [
                None,
                obj.name,
                "df_row.index.get_level_values('node')[0]._hatchet_nid <= {}".format(
                    obj.val
                ),
                "isinstance(df_row.index.get_level_values('node')[0]._hatchet_nid, Real)",
            ]
        return [
            None,
            obj.name,
            self._add_aggregation_call_to_multi_idx_predicate(
                'df_row["{}"].apply(lambda elem: elem <= {})'.format(obj.prop, obj.val)
            ),
            "is_numeric_dtype(df_row['{}'])".format(obj.prop),
        ]

    def _parse_num_gte(self, obj):
        """Processes numeric greater-than-or-equal-to predicates."""
        if obj.prop == "depth":
            if obj.val < 0:
                warnings.warn(
                    """
                    The 'depth' property of a Node is strictly non-negative.
                    This condition will always be true.
                    The statement that triggered this warning is:
                    {}
                    """.format(
                        obj
                    ),
                    RedundantQueryFilterWarning,
                )
                return [
                    None,
                    obj.name,
                    "True",
                    "isinstance(df_row.name._depth, Real)",
                ]
            return [
                None,
                obj.name,
                "df_row.name._depth >= {}".format(obj.val),
                "isinstance(df_row.name._depth, Real)",
            ]
        if obj.prop == "node_id":
            if obj.val < 0:
                warnings.warn(
                    """
                    The 'node_id' property of a Node is strictly non-negative.
                    This condition will always be true.
                    The statement that triggered this warning is:
                    {}
                    """.format(
                        obj
                    ),
                    RedundantQueryFilterWarning,
                )
                return [
                    None,
                    obj.name,
                    "True",
                    "isinstance(df_row.name._hatchet_nid, Real)",
                ]
            return [
                None,
                obj.name,
                "df_row.name._hatchet_nid >= {}".format(obj.val),
                "isinstance(df_row.name._hatchet_nid, Real)",
            ]
        return [
            None,
            obj.name,
            'df_row["{}"] >= {}'.format(obj.prop, obj.val),
            "isinstance(df_row['{}'], Real)".format(obj.prop),
        ]

    def _parse_num_gte_multi_idx(self, obj):
        if obj.prop == "depth":
            if obj.val < 0:
                warnings.warn(
                    """
                    The 'depth' property of a Node is strictly non-negative.
                    This condition will always be true.
                    The statement that triggered this warning is:
                    {}
                    """.format(
                        obj
                    ),
                    RedundantQueryFilterWarning,
                )
                return [
                    None,
                    obj.name,
                    "True",
                    "isinstance(df_row.index.get_level_values('node')[0]._depth, Real)",
                ]
            return [
                None,
                obj.name,
                "df_row.index.get_level_values('node')[0]._depth >= {}".format(obj.val),
                "isinstance(df_row.index.get_level_values('node')[0]._depth, Real)",
            ]
        if obj.prop == "node_id":
            if obj.val < 0:
                warnings.warn(
                    """
                    The 'node_id' property of a Node is strictly non-negative.
                    This condition will always be true.
                    The statement that triggered this warning is:
                    {}
                    """.format(
                        obj
                    ),
                    RedundantQueryFilterWarning,
                )
                return [
                    None,
                    obj.name,
                    "True",
                    "isinstance(df_row.index.get_level_values('node')[0]._hatchet_nid, Real)",
                ]
            return [
                None,
                obj.name,
                "df_row.index.get_level_values('node')[0]._hatchet_nid >= {}".format(
                    obj.val
                ),
                "isinstance(df_row.index.get_level_values('node')[0]._hatchet_nid, Real)",
            ]
        return [
            None,
            obj.name,
            self._add_aggregation_call_to_multi_idx_predicate(
                'df_row["{}"].apply(lambda elem: elem >= {})'.format(obj.prop, obj.val)
            ),
            "is_numeric_dtype(df_row['{}'])".format(obj.prop),
        ]

    def _parse_num_nan(self, obj):
        """Processes predicates that check for NaN."""
        if obj.prop == "depth":
            return [
                None,
                obj.name,
                "pd.isna(df_row.name._depth)",
                "isinstance(df_row.name._depth, Real)",
            ]
        if obj.prop == "node_id":
            return [
                None,
                obj.name,
                "pd.isna(df_row.name._hatchet_nid)",
                "isinstance(df_row.name._hatchet_nid, Real)",
            ]
        return [
            None,
            obj.name,
            'pd.isna(df_row["{}"])'.format(obj.prop),
            "isinstance(df_row['{}'], Real)".format(obj.prop),
        ]

    def _parse_num_nan_multi_idx(self, obj):
        if obj.prop == "depth":
            return [
                None,
                obj.name,
                "pd.isna(df_row.index.get_level_values('node')[0]._depth)",
                "isinstance(df_row.index.get_level_values('node')[0]._depth, Real)",
            ]
        if obj.prop == "node_id":
            return [
                None,
                obj.name,
                "pd.isna(df_row.index.get_level_values('node')[0]._hatchet_nid)",
                "isinstance(df_row.index.get_level_values('node')[0]._hatchet_nid, Real)",
            ]
        return [
            None,
            obj.name,
            self._add_aggregation_call_to_multi_idx_predicate(
                'pd.isna(df_row["{}"])'.format(obj.prop)
            ),
            "is_numeric_dtype(df_row['{}'])".format(obj.prop),
        ]

    def _parse_num_not_nan(self, obj):
        """Processes predicates that check for NaN."""
        if obj.prop == "depth":
            return [
                None,
                obj.name,
                "not pd.isna(df_row.name._depth)",
                "isinstance(df_row.name._depth, Real)",
            ]
        if obj.prop == "node_id":
            return [
                None,
                obj.name,
                "not pd.isna(df_row.name._hatchet_nid)",
                "isinstance(df_row.name._hatchet_nid, Real)",
            ]
        return [
            None,
            obj.name,
            'not pd.isna(df_row["{}"])'.format(obj.prop),
            "isinstance(df_row['{}'], Real)".format(obj.prop),
        ]

    def _parse_num_not_nan_multi_idx(self, obj):
        if obj.prop == "depth":
            return [
                None,
                obj.name,
                "not pd.isna(df_row.index.get_level_values('node')[0]._depth)",
                "isinstance(df_row.index.get_level_values('node')[0]._depth, Real)",
            ]
        if obj.prop == "node_id":
            return [
                None,
                obj.name,
                "not pd.isna(df_row.index.get_level_values('node')[0]._hatchet_nid)",
                "isinstance(df_row.index.get_level_values('node')[0]._hatchet_nid, Real)",
            ]
        return [
            None,
            obj.name,
            self._add_aggregation_call_to_multi_idx_predicate(
                'not pd.isna(df_row["{}"])'.format(obj.prop)
            ),
            "is_numeric_dtype(df_row['{}'])".format(obj.prop),
        ]

    def _parse_num_inf(self, obj):
        """Processes predicates that check for Infinity."""
        if obj.prop == "depth":
            return [
                None,
                obj.name,
                "np.isinf(df_row.name._depth)",
                "isinstance(df_row.name._depth, Real)",
            ]
        if obj.prop == "node_id":
            return [
                None,
                obj.name,
                "np.isinf(df_row.name._hatchet_nid)",
                "isinstance(df_row.name._hatchet_nid, Real)",
            ]
        return [
            None,
            obj.name,
            'np.isinf(df_row["{}"])'.format(obj.prop),
            "isinstance(df_row['{}'], Real)".format(obj.prop),
        ]

    def _parse_num_inf_multi_idx(self, obj):
        if obj.prop == "depth":
            return [
                None,
                obj.name,
                "np.isinf(df_row.index.get_level_values('node')[0]._depth)",
                "isinstance(df_row.index.get_level_values('node')[0]._depth, Real)",
            ]
        if obj.prop == "node_id":
            return [
                None,
                obj.name,
                "np.isinf(df_row.index.get_level_values('node')[0]._hatchet_nid)",
                "isinstance(df_row.index.get_level_values('node')[0]._hatchet_nid, Real)",
            ]
        return [
            None,
            obj.name,
            self._add_aggregation_call_to_multi_idx_predicate(
                'np.isinf(df_row["{}"])'.format(obj.prop)
            ),
            "is_numeric_dtype(df_row['{}'])".format(obj.prop),
        ]

    def _parse_num_not_inf(self, obj):
        """Processes predicates that check for not-Infinity."""
        if obj.prop == "depth":
            return [
                None,
                obj.name,
                "not np.isinf(df_row.name._depth)",
                "isinstance(df_row.name._depth, Real)",
            ]
        if obj.prop == "node_id":
            return [
                None,
                obj.name,
                "not np.isinf(df_row.name._hatchet_nid)",
                "isinstance(df_row.name._hatchet_nid, Real)",
            ]
        return [
            None,
            obj.name,
            'not np.isinf(df_row["{}"])'.format(obj.prop),
            "isinstance(df_row['{}'], Real)".format(obj.prop),
        ]

    def _parse_num_not_inf_multi_idx(self, obj):
        if obj.prop == "depth":
            return [
                None,
                obj.name,
                "not np.isinf(df_row.index.get_level_values('node')[0]._depth)",
                "isinstance(df_row.index.get_level_values('node')[0]._depth, Real)",
            ]
        if obj.prop == "node_id":
            return [
                None,
                obj.name,
                "not np.isinf(df_row.index.get_level_values('node')[0]._hatchet_nid)",
                "isinstance(df_row.index.get_level_values('node')[0]._hatchet_nid, Real)",
            ]
        return [
            None,
            obj.name,
            self._add_aggregation_call_to_multi_idx_predicate(
                'not np.isinf(df_row["{}"])'.format(obj.prop)
            ),
            "is_numeric_dtype(df_row['{}'])".format(obj.prop),
        ]


def parse_string_dialect(query_str, multi_index_mode="off"):
    """Parse all types of String-based queries, including multi-queries that leverage
    the curly brace delimiters.

    Arguments:
        query_str (str): the String-based query to be parsed

    Returns:
        (Query or CompoundQuery): A Hatchet query object representing the String-based query
    """
    # TODO Check if there's a way to prevent curly braces in a string
    #      from being captured

    # Find the number of curly brace-delimited regions in the query
    query_str = query_str.strip()
    curly_brace_elems = re.findall(r"\{(.*?)\}", query_str)
    num_curly_brace_elems = len(curly_brace_elems)
    # If there are no curly brace-delimited regions, just pass the query
    # off to the CypherQuery constructor
    if num_curly_brace_elems == 0:
        if sys.version_info[0] == 2:
            query_str = query_str.decode("utf-8")
        return StringQuery(query_str, multi_index_mode)
    # Create an iterator over the curly brace-delimited regions
    curly_brace_iter = re.finditer(r"\{(.*?)\}", query_str)
    # Will store curly brace-delimited regions in the WHERE clause
    condition_list = None
    # Will store curly brace-delimited regions that contain entire
    # mid-level queries (MATCH clause and WHERE clause)
    query_list = None
    # If entire queries are in brace-delimited regions, store the indexes
    # of the regions here so we don't consider brace-delimited regions
    # within the already-captured region.
    query_idxes = None
    # Store which compound queries to apply to the curly brace-delimited regions
    compound_ops = []
    for i, match in enumerate(curly_brace_iter):
        # Get the substring within curly braces
        substr = query_str[match.start() + 1 : match.end() - 1]
        substr = substr.strip()
        # If an entire query (MATCH + WHERE) is within curly braces,
        # add the query to "query_list", and add the indexes corresponding
        # to the query to "query_idxes"
        if substr.startswith("MATCH"):
            if query_list is None:
                query_list = []
            if query_idxes is None:
                query_idxes = []
            query_list.append(substr)
            query_idxes.append((match.start(), match.end()))
        # If the curly brace-delimited region contains only parts of a
        # WHERE clause, first, check if the region is within another
        # curly brace delimited region. If it is, do nothing (it will
        # be handled later). Otherwise, add the region to "condition_list"
        elif re.match(r"[a-zA-Z0-9_]+\..*", substr) is not None:
            is_encapsulated_region = False
            if query_idxes is not None:
                for s, e in query_idxes:
                    if match.start() >= s or match.end() <= e:
                        is_encapsulated_region = True
                        break
            if is_encapsulated_region:
                continue
            if condition_list is None:
                condition_list = []
            condition_list.append(substr)
        # If the curly brace-delimited region is neither a whole query
        # or part of a WHERE clause, raise an error
        else:
            raise ValueError("Invalid grouping (with curly braces) within the query")
        # If there is a compound operator directly after the curly brace-delimited region,
        # capture the type of operator, and store the type in "compound_ops"
        if i + 1 < num_curly_brace_elems:
            rest_substr = query_str[match.end() :]
            rest_substr = rest_substr.strip()
            if rest_substr.startswith("AND"):
                compound_ops.append("AND")
            elif rest_substr.startswith("OR"):
                compound_ops.append("OR")
            elif rest_substr.startswith("XOR"):
                compound_ops.append("XOR")
            else:
                raise ValueError("Invalid compound operator type found!")
    # Each call to this function should only consider one of the full query or
    # WHERE clause versions at a time. If both types were captured, raise an error
    # because some type of internal logic issue occured.
    if condition_list is not None and query_list is not None:
        raise ValueError(
            "Curly braces must be around either a full mid-level query or a set of conditions in a single mid-level query"
        )
    # This branch is for the WHERE clause version
    if condition_list is not None:
        # Make sure you correctly gathered curly brace-delimited regions and
        # compound operators
        if len(condition_list) != len(compound_ops) + 1:
            raise ValueError(
                "Incompatible number of curly brace elements and compound operators"
            )
        # Get the MATCH clause that will be shared across the subqueries
        match_comp_obj = re.search(r"MATCH\s+(?P<match_field>.*)\s+WHERE", query_str)
        match_comp = match_comp_obj.group("match_field")
        # Iterate over the compound operators
        full_query = None
        for i, op in enumerate(compound_ops):
            # If in the first iteration, set the initial query as a CypherQuery where
            # the MATCH clause is the shared match clause and the WHERE clause is the
            # first curly brace-delimited region
            if i == 0:
                query1 = "MATCH {} WHERE {}".format(match_comp, condition_list[i])
                if sys.version_info[0] == 2:
                    query1 = query1.decode("utf-8")
                full_query = StringQuery(query1, multi_index_mode)
            # Get the next query as a CypherQuery where
            # the MATCH clause is the shared match clause and the WHERE clause is the
            # next curly brace-delimited region
            next_query = "MATCH {} WHERE {}".format(match_comp, condition_list[i + 1])
            if sys.version_info[0] == 2:
                next_query = next_query.decode("utf-8")
            next_query = StringQuery(next_query, multi_index_mode)
            # Add the next query to the full query using the compound operator
            # currently being considered
            if op == "AND":
                full_query = full_query & next_query
            elif op == "OR":
                full_query = full_query | next_query
            else:
                full_query = full_query ^ next_query
        return full_query
    # This branch is for the full query version
    else:
        # Make sure you correctly gathered curly brace-delimited regions and
        # compound operators
        if len(query_list) != len(compound_ops) + 1:
            raise ValueError(
                "Incompatible number of curly brace elements and compound operators"
            )
        # Iterate over the compound operators
        full_query = None
        for i, op in enumerate(compound_ops):
            # If in the first iteration, set the initial query as the result
            # of recursively calling this function on the first curly brace-delimited region
            if i == 0:
                full_query = parse_string_dialect(query_list[i])
            # Get the next query by recursively calling this function
            # on the next curly brace-delimited region
            next_query = parse_string_dialect(query_list[i + 1])
            # Add the next query to the full query using the compound operator
            # currently being considered
            if op == "AND":
                full_query = full_query & next_query
            elif op == "OR":
                full_query = full_query | next_query
            else:
                full_query = full_query ^ next_query
        return full_query



# Copyright 2017-2023 Lawrence Livermore National Security, LLC and other
# Hatchet Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

import matplotlib.cm
import matplotlib.colors


def trees_to_dot(roots, dataframe, metric, name, rank, thread, threshold):
    """Calls to_dot in turn for each tree in the graph/forest."""
    text = (
        "strict digraph {\n"
        "graph [bgcolor=transparent];\n"
        "node [penwidth=4, shape=circle];\n"
        "edge [penwidth=2];\n\n"
    )

    all_nodes = ""
    all_edges = ""

    # call to_dot for each root in the graph
    visited = []
    for root in roots:
        (nodes, edges) = to_dot(
            root, dataframe, metric, name, rank, thread, threshold, visited
        )
        all_nodes += nodes
        all_edges += edges

    text += all_nodes + "\n" + all_edges + "\n}\n"

    return text


def to_dot(hnode, dataframe, metric, name, rank, thread, threshold, visited):
    """Write to graphviz dot format."""
    colormap = matplotlib.cm.Reds
    min_time = dataframe[metric].min()
    max_time = dataframe[metric].max()

    def add_nodes_and_edges(hnode):

        # set dataframe index based on if rank is a part of the index
        if "rank" in dataframe.index.names and "thread" in dataframe.index.names:
            df_index = (hnode, rank, thread)
        elif "rank" in dataframe.index.names:
            df_index = (hnode, rank)
        elif "thread" in dataframe.index.names:
            df_index = (hnode, thread)
        else:
            df_index = hnode

        node_time = dataframe.loc[df_index, metric]
        node_name = dataframe.loc[df_index, name]
        node_id = hnode._hatchet_nid

        weight = (node_time - min_time) / (max_time - min_time)
        color = matplotlib.colors.rgb2hex(colormap(weight))

        # only display nodes whose metric is greater than some threshold
        if (node_time >= threshold * max_time) and (hnode not in visited):
            node_string = '"{0}" [color="{1}", label="{2}" shape=oval];\n'.format(
                node_id, color, node_name
            )
            edge_string = ""

            # only display those edges where child's metric is greater than
            # threshold
            children = []
            for child in hnode.children:
                if (
                    "rank" in dataframe.index.names
                    and "thread" in dataframe.index.names
                ):
                    df_index = (child, rank, thread)
                elif "rank" in dataframe.index.names:
                    df_index = (child, rank)
                elif "thread" in dataframe.index.names:
                    df_index = (child, thread)
                else:
                    df_index = child

                child_time = dataframe.loc[df_index, metric]
                if child_time >= threshold * max_time:
                    children.append(child)

            visited.append(hnode)
            for child in children:
                # add edges
                child_id = child._hatchet_nid

                edge_string += '"{0}" -> "{1}";\n'.format(node_id, child_id)
                (nodes, edges) = add_nodes_and_edges(child)
                node_string += nodes
                edge_string += edges
        else:
            node_string = ""
            edge_string = ""

        return (node_string, edge_string)

    # call add_nodes_and_edges on the root
    return add_nodes_and_edges(hnode)



# Copyright 2017-2023 Lawrence Livermore National Security, LLC and other
# Hatchet Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

from hatchet.node import Node

from abc import abstractmethod

# TODO The ABC class was introduced in Python 3.4.
# When support for earlier versions is (eventually) dropped,
# this entire "try-except" block can be reduced to:
# from abc import ABC
try:
    from abc import ABC
except ImportError:
    from abc import ABCMeta

    ABC = ABCMeta("ABC", (object,), {"__slots__": ()})


def _get_node_from_df_iloc(df, ind):
    node = None
    if isinstance(df.iloc[ind].name, tuple):
        node = df.iloc[ind].name[0]
    elif isinstance(df.iloc[ind].name, Node):
        node = df.iloc[ind].name
    else:
        raise InvalidDataFrameIndex(
            "DataFrame index elements must be either a tuple or a Node"
        )
    return node


def _fill_children_and_parents(dump_df):
    dump_df["children"] = [[] for _ in range(len(dump_df))]
    dump_df["parents"] = [[] for _ in range(len(dump_df))]
    for i in range(len(dump_df)):
        node = _get_node_from_df_iloc(dump_df, i)
        dump_df.iat[i, dump_df.columns.get_loc("children")] = [
            c._hatchet_nid for c in node.children
        ]
        node.children = []
        dump_df.iat[i, dump_df.columns.get_loc("parents")] = [
            p._hatchet_nid for p in node.parents
        ]
        node.parents = []
    return dump_df


class DataframeWriter(ABC):
    def __init__(self, filename):
        self.filename = filename

    @abstractmethod
    def _write_dataframe_to_file(self, df, **kwargs):
        pass

    def write(self, gf, **kwargs):
        gf_cpy = gf.deepcopy()
        dump_df = _fill_children_and_parents(gf_cpy.dataframe)
        dump_df["exc_metrics"] = None
        dump_df.iat[0, dump_df.columns.get_loc("exc_metrics")] = gf_cpy.exc_metrics
        dump_df["inc_metrics"] = None
        dump_df.iat[0, dump_df.columns.get_loc("inc_metrics")] = gf_cpy.inc_metrics
        dump_df["default_metric"] = None
        dump_df.iat[
            0, dump_df.columns.get_loc("default_metric")
        ] = gf_cpy.default_metric
        self._write_dataframe_to_file(dump_df, **kwargs)


class InvalidDataFrameIndex(Exception):
    """Raised when the DataFrame index is of an invalid type."""



# Copyright 2017-2023 Lawrence Livermore National Security, LLC and other
# Hatchet Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

from itertools import groupby
import pandas as pd

from .errors import InvalidQueryFilter
from ..node import Node, traversal_order
from .query import Query
from .compound import CompoundQuery
from .object_dialect import ObjectQuery
from .string_dialect import parse_string_dialect


class QueryEngine:

    """Class for applying queries to GraphFrames."""

    def __init__(self):
        """Creates the QueryEngine."""
        self.search_cache = {}

    def reset_cache(self):
        """Resets the cache in the QueryEngine."""
        self.search_cache = {}

    def apply(self, query, graph, dframe):
        """Apply the query to a GraphFrame.

        Arguments:
            query (Query or CompoundQuery): the query being applied
            graph (Graph): the Graph to which the query is being applied
            dframe (pandas.DataFrame): the DataFrame associated with the graph

        Returns:
            (list): A list representing the set of nodes from paths that match the query
        """
        if issubclass(type(query), Query):
            self.reset_cache()
            matches = []
            visited = set()
            for root in sorted(graph.roots, key=traversal_order):
                self._apply_impl(query, dframe, root, visited, matches)
            assert len(visited) == len(graph)
            matched_node_set = list(set().union(*matches))
            # return matches
            return matched_node_set
        elif issubclass(type(query), CompoundQuery):
            results = []
            for subq in query.subqueries:
                subq_obj = subq
                if isinstance(subq, list):
                    subq_obj = ObjectQuery(subq)
                elif isinstance(subq, str):
                    subq_obj = parse_string_dialect(subq)
                results.append(self.apply(subq_obj, graph, dframe))
            return query._apply_op_to_results(results, graph)
        else:
            raise TypeError("Invalid query data type ({})".format(str(type(query))))

    def _cache_node(self, node, query, dframe):
        """Cache (Memoize) the parts of the query that the node matches.

        Arguments:
            node (Node): the Node to be cached
            query (Query): the query being applied
            dframe (pandas.DataFrame): the DataFrame containing node metrics and other data
        """
        assert isinstance(node, Node)
        matches = []
        # Applies each filtering function to the node to cache which
        # query nodes the current node matches.
        for i, node_query in enumerate(query):
            _, filter_func = node_query
            row = None
            if isinstance(dframe.index, pd.MultiIndex):
                row = pd.concat([dframe.loc[node]], keys=[node], names=["node"])
            else:
                row = dframe.loc[node]
            if filter_func(row):
                matches.append(i)
        self.search_cache[node._hatchet_nid] = matches

    def _match_0_or_more(self, query, dframe, node, wcard_idx):
        """Process a "*" predicate in the query on a subgraph.

        Arguments:
            query (Query): the query being applied
            dframe (pandas.DataFrame): the DataFrame containing the metrics for the queried GraphFrame
            node (Node): the node being queried against the "*" predicate
            wcard_idx (int): the index into the query associated with the "*" predicate

        Returns:
            (list): a list of lists representing the paths rooted at "node" that match the "*" predicate
                    and/or the next query node. Will return None if there is no match for the "*"
                    predicate or the next query node.
        """
        # Cache the node if it's not already cached
        if node._hatchet_nid not in self.search_cache:
            self._cache_node(node, query, dframe)
        # If the node matches with the next non-wildcard query node,
        # end the recursion and return the node.
        if wcard_idx + 1 in self.search_cache[node._hatchet_nid]:
            return [[]]
        # If the node matches the "*" wildcard query, recursively
        # apply this function to the current node's children. Then,
        # collect their returned matches, and prepend the current node.
        elif wcard_idx in self.search_cache[node._hatchet_nid]:
            matches = []
            if len(node.children) == 0:
                if wcard_idx == len(query) - 1:
                    return [[node]]
                return None
            for child in sorted(node.children, key=traversal_order):
                sub_match = self._match_0_or_more(query, dframe, child, wcard_idx)
                if sub_match is not None:
                    matches.extend(sub_match)
            if len(matches) == 0:
                return None
            tmp = set(tuple(m) for m in matches)
            matches = [list(t) for t in tmp]
            return [[node] + m for m in matches]
        # If the current node doesn't match the current "*" wildcard or
        # the next non-wildcard query node, return None.
        else:
            if wcard_idx == len(query) - 1:
                return [[]]
            return None

    def _match_1(self, query, dframe, node, idx):
        """Process a "." predicate in the query on a subgraph.

        Arguments:
            query (Query): the query being applied
            dframe (pandas.DataFrame): the DataFrame containing the metrics for the queried GraphFrame
            node (Node): the node being queried against the "." predicate
            idx (int): the index into the query associated with the "." predicate

        Returns:
            (list): A list of lists representing the children of "node" that match the "." predicate being considered.
                    Will return None if there are no matches for the "." predicate.
        """
        if node._hatchet_nid not in self.search_cache:
            self._cache_node(node, query, dframe)
        matches = []
        for child in sorted(node.children, key=traversal_order):
            # Cache the node if it's not already cached
            if child._hatchet_nid not in self.search_cache:
                self._cache_node(child, query, dframe)
            if idx in self.search_cache[child._hatchet_nid]:
                matches.append([child])
        # To be consistent with the other matching functions, return
        # None instead of an empty list.
        if len(matches) == 0:
            return None
        return matches

    def _match_pattern(self, query, dframe, pattern_root, match_idx):
        """Try to match the query pattern starting at the provided root node.

        Arguments:
            query (Query): the query being applied
            dframe (pandas.DataFrame): the DataFrame containing the metrics for the queried GraphFrame
            pattern_root (Node): the current node considered in the query
            match_idx (int): the current index into the query

        Returns:
            (list): A list of lists representing the paths rooted at "pattern_root" that match the query
        """
        assert isinstance(pattern_root, Node)
        # Starting query node
        pattern_idx = match_idx + 1
        if query.query_pattern[match_idx][0] == "*":
            pattern_idx = 0
        # Starting matching pattern
        matches = [[pattern_root]]
        while pattern_idx < len(query):
            # Get the wildcard type
            wcard, _ = query.query_pattern[pattern_idx]
            new_matches = []
            # Consider each existing match individually so that more
            # nodes can be added to them.
            for m in matches:
                sub_match = []
                # Get the portion of the subgraph that matches the next
                # part of the query.
                if wcard == ".":
                    s = self._match_1(query, dframe, m[-1], pattern_idx)
                    if s is None:
                        sub_match.append(s)
                    else:
                        sub_match.extend(s)
                elif wcard == "*":
                    if len(m[-1].children) == 0:
                        sub_match.append([])
                    else:
                        for child in sorted(m[-1].children, key=traversal_order):
                            s = self._match_0_or_more(query, dframe, child, pattern_idx)
                            if s is None:
                                sub_match.append(s)
                            else:
                                sub_match.extend(s)
                else:
                    raise InvalidQueryFilter(
                        'Query wildcards must (internally) be one of "." or "*"'
                    )
                # Merge the next part of the match path with the
                # existing part.
                for s in sub_match:
                    if s is not None:
                        new_matches.append(m + s)
                new_matches = [uniq_match for uniq_match, _ in groupby(new_matches)]
            # Overwrite the old matches with the updated matches
            matches = new_matches
            # If all the existing partial matches were not able to be
            # expanded into full matches, return None.
            if len(matches) == 0:
                return None
            # Update the query node
            pattern_idx += 1
        return matches

    def _apply_impl(self, query, dframe, node, visited, matches):
        """Traverse the subgraph with the specified root, and collect all paths that match the query.

        Arguments:
            query (Query): the query being applied
            dframe (pandas.DataFrame): the DataFrame containing the metrics for the queried GraphFrame
            node (Node): the root node of the subgraph that is being queried
            visited (set): a set that keeps track of what nodes have been visited in the traversal to minimize the amount of work that is repeated
            matches (list): the list in which the final set of matches are stored
        """
        # If the node has already been visited (or is None for some
        # reason), skip it.
        if node is None or node._hatchet_nid in visited:
            return
        # Cache the node if it's not already cached
        if node._hatchet_nid not in self.search_cache:
            self._cache_node(node, query, dframe)
        # If the node matches the starting/root node of the query,
        # try to get all query matches in the subgraph rooted at
        # this node.
        if query.query_pattern[0][0] == "*":
            if 1 in self.search_cache[node._hatchet_nid]:
                sub_match = self._match_pattern(query, dframe, node, 1)
                if sub_match is not None:
                    matches.extend(sub_match)
        if 0 in self.search_cache[node._hatchet_nid]:
            sub_match = self._match_pattern(query, dframe, node, 0)
            if sub_match is not None:
                matches.extend(sub_match)
        # Note that the node is now visited.
        visited.add(node._hatchet_nid)
        # Continue the Depth First Search.
        for child in sorted(node.children, key=traversal_order):
            self._apply_impl(query, dframe, child, visited, matches)



# Copyright 2017-2023 Lawrence Livermore National Security, LLC and other
# Hatchet Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

import functools


def deprecated_params(**old_to_new):
    def deco(f):
        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            rename_kwargs(f.__name__, old_to_new, kwargs)
            return f(*args, **kwargs)

        return wrapper

    return deco


def rename_kwargs(fname, old_to_new, kwargs):
    for old, new in old_to_new.items():
        if old in kwargs:
            if new in kwargs:
                raise TypeError(
                    '{}() received both "{}=" and "{}=".'.format(fname, old, new)
                )

            # if parameter has been removed
            if not new:
                raise ValueError(
                    '{}() parameter "{}=" will be removed in a future release. Please remove it from this script.'.format(
                        fname, old
                    )
                )
            # if parameter has been deprecated and renamed
            else:
                raise ValueError(
                    '{}() parameter "{}=" has been deprecated, please use "{}=".'.format(
                        fname, old, new
                    )
                )
                kwargs[new] = kwargs.pop(old)



# Copyright 2017-2023 Lawrence Livermore National Security, LLC and other
# Hatchet Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

from IPython.core.magic import Magics, magics_class, line_magic
from hatchet.external import Roundtrip as RT
from hatchet import GraphFrame
from os import path
from os.path import dirname


vis_dir = dirname(path.abspath(__file__))


def _gf_to_json(data):
    import json
    from pandas import Series

    def serialize(obj):
        if isinstance(obj, Series):
            return obj.to_json
        return obj.__dict__

    try:
        if isinstance(data, GraphFrame):
            return json.dumps(data.to_literal(), default=serialize)
        else:
            with open("check", "w") as f:
                f.write(json.dumps(data, default=serialize))
            return json.dumps(data, default=serialize)
    except ValueError:
        raise "Input data is not of type graphframe or json serializable."


def _query_to_dict(json_query):
    import json

    return json_query

    return json.loads(json_query)


@magics_class
class CCT(Magics):
    def __init__(self, shell):
        super(CCT, self).__init__(shell)
        self.vis_dist = path.join(vis_dir, "static")

    @line_magic
    def cct(self, line):
        args = line.split(" ")

        RT.load_webpack(path.join(self.vis_dist, "cct_bundle.html"), cache=False)
        RT.var_to_js(
            args[0], "hatchet_tree_def", watch=False, to_js_converter=_gf_to_json
        )

        RT.initialize()

    @line_magic
    def cct_fetch_query(self, line):
        args = line.split(" ")

        RT.fetch_data("jsNodeSelected", args[0], converter=_query_to_dict)


def load_ipython_extension(ipython):
    ipython.register_magics(CCT)



# Copyright 2017-2023 Lawrence Livermore National Security, LLC and other
# Hatchet Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

import json

import pandas as pd

import hatchet.graphframe
from hatchet.node import Node
from hatchet.graph import Graph
from hatchet.frame import Frame


class PyinstrumentReader:
    def __init__(self, filename):
        self.pyinstrument_json_filename = filename
        self.graph_dict = {}
        self.list_roots = []
        self.node_dicts = []

    def create_graph(self):
        def parse_node_literal(child_dict, hparent):
            """Create node_dict for one node and then call the function
            recursively on all children."""

            hnode = Node(
                Frame({"name": child_dict["function"], "type": "function"}), hparent
            )

            child_node_dict = {
                "node": hnode,
                "name": child_dict["function"],
                "file": child_dict["file_path_short"],
                "line": child_dict["line_no"],
                "time": child_dict["time"],
                "time (inc)": child_dict["time"],
                "is_application_code": child_dict["is_application_code"],
            }

            hparent.add_child(hnode)
            self.node_dicts.append(child_node_dict)

            if "children" in child_dict:
                for child in child_dict["children"]:
                    # Pyinstrument's time metric actually stores inclusive time.
                    # To calculate exclusive time, we subtract the children's time
                    # from the parent's time.
                    child_node_dict["time"] -= child["time"]
                    parse_node_literal(child, hnode)

        # start with creating a node_dict for each root
        graph_root = Node(
            Frame(
                {"name": self.graph_dict["root_frame"]["function"], "type": "function"}
            ),
            None,
        )

        node_dict = {
            "node": graph_root,
            "name": self.graph_dict["root_frame"]["function"],
            "file": self.graph_dict["root_frame"]["file_path_short"],
            "line": self.graph_dict["root_frame"]["line_no"],
            "time": self.graph_dict["root_frame"]["time"],
            "time (inc)": self.graph_dict["root_frame"]["time"],
            "is_application_code": self.graph_dict["root_frame"]["is_application_code"],
        }

        self.node_dicts.append(node_dict)
        self.list_roots.append(graph_root)

        # call recursively on all children of root
        if "children" in self.graph_dict["root_frame"]:
            for child in self.graph_dict["root_frame"]["children"]:
                # Pyinstrument's time metric actually stores inclusive time.
                # To calculate exclusive time, we subtract the children's time
                # from the parent's time.
                node_dict["time"] -= child["time"]
                parse_node_literal(child, graph_root)

        graph = Graph(self.list_roots)
        graph.enumerate_traverse()

        return graph

    def read(self):
        with open(self.pyinstrument_json_filename) as pyinstrument_json:
            self.graph_dict = json.load(pyinstrument_json)

        graph = self.create_graph()

        dataframe = pd.DataFrame(data=self.node_dicts)
        dataframe.set_index(["node"], inplace=True)
        dataframe.sort_index(inplace=True)

        return hatchet.graphframe.GraphFrame(graph, dataframe, ["time"], ["time (inc)"])



# Copyright 2017-2023 Lawrence Livermore National Security, LLC and other
# Hatchet Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

__version_info__ = ("2023", "1", "0")
__version__ = ".".join(__version_info__)



# Copyright 2017-2023 Lawrence Livermore National Security, LLC and other
# Hatchet Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

# Make flake8 ignore unused names in this file
# flake8: noqa: F401

from .query import Query
from .compound import (
    CompoundQuery,
    ConjunctionQuery,
    DisjunctionQuery,
    ExclusiveDisjunctionQuery,
    NegationQuery,
)
from .object_dialect import ObjectQuery
from .string_dialect import StringQuery, parse_string_dialect
from .engine import QueryEngine
from .errors import (
    InvalidQueryPath,
    InvalidQueryFilter,
    RedundantQueryFilterWarning,
    BadNumberNaryQueryArgs,
)

from .compat import (
    AbstractQuery,
    NaryQuery,
    AndQuery,
    IntersectionQuery,
    OrQuery,
    UnionQuery,
    XorQuery,
    SymDifferenceQuery,
    NotQuery,
    QueryMatcher,
    CypherQuery,
    parse_cypher_query,
)


def combine_via_conjunction(query0, query1):
    return ConjunctionQuery(query0, query1)


def combine_via_disjunction(query0, query1):
    return DisjunctionQuery(query0, query1)


def combine_via_exclusive_disjunction(query0, query1):
    return ExclusiveDisjunctionQuery(query0, query1)


def negate_query(query):
    return NegationQuery(query)


Query.__and__ = combine_via_conjunction
Query.__or__ = combine_via_disjunction
Query.__xor__ = combine_via_exclusive_disjunction
Query.__not__ = negate_query


CompoundQuery.__and__ = combine_via_conjunction
CompoundQuery.__or__ = combine_via_disjunction
CompoundQuery.__xor__ = combine_via_exclusive_disjunction
CompoundQuery.__not__ = negate_query


def is_hatchet_query(query_obj):
    return (
        issubclass(type(query_obj), Query)
        or issubclass(type(query_obj), CompoundQuery)
        or issubclass(type(query_obj), AbstractQuery)
    )


# Uncomment when accessing old-style queries using
# 'from hatchet.query import *' is removed
#
# __all__ = [
#     "Query",
#     "CompoundQuery",
#     "ConjunctionQuery",
#     "DisjunctionQuery",
#     "ExclusiveDisjunctionQuery",
#     "NegationQuery",
#     "ObjectQuery",
#     "StringQuery",
#     "parse_string_dialect",
#     "InvalidQueryFilter",
#     "InvalidQueryPath",
#     "RedundantQueryFilterWarning",
#     "BadNumberNaryQueryArgs",
# ]



# -*- coding: utf-8 -*-

# Copyright (c) 2014-2018, Joe Rickerby and contributors
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from ..version import __version__

import pandas as pd
import numpy as np
import warnings
from ..util.colormaps import ColorMaps


class ConsoleRenderer:
    def __init__(self, unicode=False, color=False):
        self.unicode = unicode
        self.color = color
        self.visited = []

    def render(self, roots, dataframe, **kwargs):
        self.render_header = kwargs["render_header"]

        if self.render_header:
            result = self.render_preamble()
        else:
            result = ""

        if roots is None:
            result += "The graph is empty.\n\n"
            return result

        self.metric_columns = kwargs["metric_column"]
        self.precision = kwargs["precision"]
        self.name = kwargs["name_column"]
        self.expand = kwargs["expand_name"]
        self.context = kwargs["context_column"]
        self.rank = kwargs["rank"]
        self.thread = kwargs["thread"]
        self.depth = kwargs["depth"]
        self.highlight = kwargs["highlight_name"]
        self.colormap = kwargs["colormap"]
        self.invert_colormap = kwargs["invert_colormap"]

        if self.color:
            self.colors = self.colors_enabled
            # set the colormap based on user input
            self.colors.colormap = ColorMaps().get_colors(
                self.colormap, self.invert_colormap
            )
        else:
            self.colors = self.colors_disabled

        if isinstance(self.metric_columns, str):
            self.primary_metric = self.metric_columns
            self.second_metric = None
        elif isinstance(self.metric_columns, list):
            if len(self.metric_columns) > 2:
                warnings.warn(
                    "More than 2 metrics specified in metric_column=. Tree() will only show 2 metrics at a time. The remaining metrics will not be shown.",
                    SyntaxWarning,
                )
                self.primary_metric = self.metric_columns[0]
                self.second_metric = self.metric_columns[1]
            elif len(self.metric_columns) == 2:
                self.primary_metric = self.metric_columns[0]
                self.second_metric = self.metric_columns[1]
            elif len(self.metric_columns) == 1:
                self.primary_metric = self.metric_columns[0]
                self.second_metric = None

        if self.primary_metric not in dataframe.columns:
            raise KeyError(
                "metric_column={} does not exist in the dataframe, please select a valid column. See a list of the available metrics with GraphFrame.show_metric_columns().".format(
                    self.primary_metric
                )
            )
        if (
            self.second_metric is not None
            and self.second_metric not in dataframe.columns
        ):
            raise KeyError(
                "metric_column={} does not exist in the dataframe, please select a valid column. See a list of the available metrics with GraphFrame.show_metric_columns().".format(
                    self.second_metric
                )
            )

        # grab the min and max value for the primary metric, ignoring inf and
        # nan values

        if "rank" in dataframe.index.names:
            metric_series = (dataframe.xs(self.rank, level=1))[self.primary_metric]
        else:
            metric_series = dataframe[self.primary_metric]
        isfinite_mask = np.isfinite(metric_series.values)
        filtered_series = pd.Series(
            metric_series.values[isfinite_mask], metric_series.index[isfinite_mask]
        )

        self.max_metric = filtered_series.max()
        self.min_metric = filtered_series.min()

        if self.unicode:
            self.lr_arrows = {"◀": u"◀ ", "▶": u"▶ "}
        else:
            self.lr_arrows = {"◀": u"< ", "▶": u"> "}

        for root in sorted(roots, key=lambda n: n.frame):
            result += self.render_frame(root, dataframe)

        if self.color is True:
            result += self.render_legend()

        if self.unicode:
            return result
        else:
            return result.encode("utf-8")

    # pylint: disable=W1401
    def render_preamble(self):
        lines = [
            r"    __          __       __         __ ",
            r"   / /_  ____ _/ /______/ /_  ___  / /_",
            r"  / __ \/ __ `/ __/ ___/ __ \/ _ \/ __/",
            r" / / / / /_/ / /_/ /__/ / / /  __/ /_  ",
            r"/_/ /_/\__,_/\__/\___/_/ /_/\___/\__/  {:>2}".format("v" + __version__),
            r"",
            r"",
        ]

        return "\n".join(lines)

    def render_legend(self):
        def render_label(index, low, high):
            metric_range = self.max_metric - self.min_metric

            return (
                self.colors.colormap[index]
                + u"█ "
                + self.colors.end
                + "{:.2f}".format(low * metric_range + self.min_metric)
                + " - "
                + "{:.2f}".format(high * metric_range + self.min_metric)
                + "\n"
            )

        legend = (
            "\n"
            + "\033[4m"
            + "Legend"
            + self.colors.end
            + " (Metric: "
            + self.primary_metric
            + " Min: {:.2f}".format(self.min_metric)
            + " Max: {:.2f}".format(self.max_metric)
            + ")\n"
        )

        legend += render_label(0, 0.9, 1.0)
        legend += render_label(1, 0.7, 0.9)
        legend += render_label(2, 0.5, 0.7)
        legend += render_label(3, 0.3, 0.5)
        legend += render_label(4, 0.1, 0.3)
        legend += render_label(5, 0.0, 0.1)

        legend += "\n" + self._ansi_color_for_name("name") + "name" + self.colors.end
        legend += " User code    "

        legend += self.colors.left + self.lr_arrows["◀"] + self.colors.end
        legend += " Only in left graph    "
        legend += self.colors.right + self.lr_arrows["▶"] + self.colors.end
        legend += " Only in right graph\n"

        return legend

    def render_frame(self, node, dataframe, indent=u"", child_indent=u""):
        node_depth = node._depth
        if node_depth < self.depth:
            # set dataframe index based on whether rank and thread are part of
            # the MultiIndex
            if "rank" in dataframe.index.names and "thread" in dataframe.index.names:
                df_index = (node, self.rank, self.thread)
            elif "rank" in dataframe.index.names:
                df_index = (node, self.rank)
            elif "thread" in dataframe.index.names:
                df_index = (node, self.thread)
            else:
                df_index = node

            node_metric = dataframe.loc[df_index, self.primary_metric]

            metric_precision = "{:." + str(self.precision) + "f}"
            metric_str = (
                self._ansi_color_for_metric(node_metric)
                + metric_precision.format(node_metric)
                + self.colors.end
            )

            if self.second_metric is not None:
                metric_str += u" {c.faint}{second_metric:.{precision}f}{c.end}".format(
                    second_metric=dataframe.loc[df_index, self.second_metric],
                    precision=self.precision,
                    c=self.colors,
                )

            node_name = dataframe.loc[df_index, self.name]
            if self.expand is False:
                if len(node_name) > 39:
                    node_name = (
                        node_name[:18] + "..." + node_name[(len(node_name) - 18) :]
                    )
            name_str = (
                self._ansi_color_for_name(node_name) + node_name + self.colors.end
            )

            # 0 is "", 1 is "L", and 2 is "R"
            if "_missing_node" in dataframe.columns:
                left_or_right = dataframe.loc[df_index, "_missing_node"]
                if left_or_right == 0:
                    lr_decorator = u""
                elif left_or_right == 1:
                    lr_decorator = u" {c.left}{decorator}{c.end}".format(
                        decorator=self.lr_arrows["◀"], c=self.colors
                    )
                elif left_or_right == 2:
                    lr_decorator = u" {c.right}{decorator}{c.end}".format(
                        decorator=self.lr_arrows["▶"], c=self.colors
                    )

            result = u"{indent}{metric_str} {name_str}".format(
                indent=indent, metric_str=metric_str, name_str=name_str
            )
            if "_missing_node" in dataframe.columns:
                result += lr_decorator
            if self.context in dataframe.columns:
                result += u" {c.faint}{context}{c.end}\n".format(
                    context=dataframe.loc[df_index, self.context], c=self.colors
                )
            else:
                result += u"\n"

            if self.unicode:
                indents = {"├": u"├─ ", "│": u"│  ", "└": u"└─ ", " ": u"   "}
            else:
                indents = {"├": u"|- ", "│": u"|  ", "└": u"`- ", " ": u"   "}

            # ensures that we never revisit nodes in the case of
            # large complex graphs
            if node not in self.visited:
                self.visited.append(node)
                sorted_children = sorted(node.children, key=lambda n: n.frame)
                if sorted_children:
                    last_child = sorted_children[-1]

                for child in sorted_children:
                    if child is not last_child:
                        c_indent = child_indent + indents["├"]
                        cc_indent = child_indent + indents["│"]
                    else:
                        c_indent = child_indent + indents["└"]
                        cc_indent = child_indent + indents[" "]
                    result += self.render_frame(
                        child, dataframe, indent=c_indent, child_indent=cc_indent
                    )
        else:
            result = ""
            indents = {"├": u"", "│": u"", "└": u"", " ": u""}

        return result

    def _ansi_color_for_metric(self, metric):
        metric_range = self.max_metric - self.min_metric

        if metric_range != 0:
            proportion_of_total = (metric - self.min_metric) / metric_range
        else:
            proportion_of_total = metric / 1

        if proportion_of_total > 0.9:
            return self.colors.colormap[0]
        elif proportion_of_total > 0.7:
            return self.colors.colormap[1]
        elif proportion_of_total > 0.5:
            return self.colors.colormap[2]
        elif proportion_of_total > 0.3:
            return self.colors.colormap[3]
        elif proportion_of_total > 0.1:
            return self.colors.colormap[4]
        elif proportion_of_total >= 0:
            return self.colors.colormap[5]
        else:
            return self.colors.blue

    def _ansi_color_for_name(self, node_name):
        if self.highlight is False:
            return ""

        if "<unknown procedure>" in node_name or "<unknown file>" in node_name:
            return ""
        else:
            return self.colors.bg_white_255 + self.colors.dark_gray_255

    class colors_enabled:
        colormap = []

        blue = "\033[34m"
        cyan = "\033[36m"

        bg_white_255 = "\033[48;5;246m"
        dark_gray_255 = "\033[38;5;232m"

        left = "\033[38;5;160m"
        right = "\033[38;5;28m"

        faint = "\033[2m"
        end = "\033[0m"

    class colors_disabled:
        colormap = ["", "", "", "", "", "", ""]

        def __getattr__(self, key):
            return ""

    colors_disabled = colors_disabled()



# Copyright 2017-2023 Lawrence Livermore National Security, LLC and other
# Hatchet Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

import pytest

from hatchet import GraphFrame
from hatchet.readers.spotdb_reader import SpotDatasetReader, SpotDBReader

spotdb = pytest.importorskip("spotdb")


def test_spot_dataset_reader():
    """Sanity-check the Spot dataset reader"""

    regionprofile = {
        "a/b/c": {"m": 20, "m#inclusive": 20},
        "a/b": {"m#inclusive": 40},
        "a": {"m#inclusive": 42},
    }
    metadata = {"launchdate": 123456789}
    attr_info = {
        "m": {"type": "double"},
        "m#inclusive": {"type": "int", "alias": "M Alias"},
    }

    reader = SpotDatasetReader(regionprofile, metadata, attr_info)
    gf = reader.read(default_metric="M Alias (inc)")

    assert len(gf.dataframe) == 3
    assert set(gf.dataframe.columns) == {"name", "m", "M Alias (inc)"}

    assert gf.metadata["launchdate"] == metadata["launchdate"]
    assert gf.default_metric == "M Alias (inc)"


@pytest.mark.skipif(not spotdb, reason="spotdb module not available")
def test_spotdb_reader(spotdb_data):
    """Sanity check for the SpotDB reader"""

    db = spotdb_data

    reader = SpotDBReader(db)
    gfs = reader.read()

    assert len(gfs) == 4

    metrics = {"Total time (inc)", "Avg time/rank (inc)"}

    assert len(gfs[0].dataframe) > 2
    assert gfs[0].default_metric == "Total time (inc)"
    assert metrics < set(gfs[0].dataframe.columns)
    assert metrics < set(gfs[3].dataframe.columns)

    assert "launchdate" in gfs[0].metadata.keys()


@pytest.mark.skipif(not spotdb, reason="spotdb module not available")
def test_from_spotdb(spotdb_data):
    """Sanity check for GraphFrame.from_spotdb"""

    db = spotdb_data
    runs = db.get_all_run_ids()
    gfs = GraphFrame.from_spotdb(spotdb_data, runs[0:2])

    assert len(gfs) == 2

    metrics = {"Total time (inc)", "Avg time/rank (inc)"}

    assert len(gfs[0].dataframe) > 2
    assert gfs[0].default_metric == "Total time (inc)"
    assert metrics < set(gfs[0].dataframe.columns)

    assert "launchdate" in gfs[0].metadata.keys()



# Copyright 2017-2023 Lawrence Livermore National Security, LLC and other
# Hatchet Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

import os


def which(executable):
    """Finds an `executable` in the user's PATH like command-line which.

    Args:
        executable (str): executable to search for
    """
    path = os.environ.get("PATH", "/usr/sbin:/usr/bin:/sbin:/bin")
    path = path.split(os.pathsep)

    for directory in path:
        exe = os.path.join(directory, executable)
        if os.path.isfile(exe) and os.access(exe, os.X_OK):
            return exe

    return None



# Copyright 2017-2023 Lawrence Livermore National Security, LLC and other
# Hatchet Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT


import pandas as pd
import os

import caliperreader as cr

import hatchet.graphframe
from hatchet.node import Node
from hatchet.graph import Graph
from hatchet.frame import Frame
from hatchet.util.timer import Timer


class CaliperNativeReader:
    """Read in a native `.cali` file using Caliper's python reader."""

    def __init__(self, filename_or_caliperreader, native, string_attributes):
        """Read in a native cali using Caliper's python reader.

        Args:
            filename_or_caliperreader (str or CaliperReader): name of a `cali` file OR
                a CaliperReader object
            native (bool): use native metric names or user-readable metric names
            string_attributes (str or list): Adds existing string attributes from within the caliper file to the dataframe
        """
        self.filename_or_caliperreader = filename_or_caliperreader
        self.filename_ext = ""
        self.use_native_metric_names = native
        self.string_attributes = string_attributes

        self.df_nodes = {}
        self.metric_cols = []
        self.record_data_cols = []
        self.node_dicts = []
        self.callpath_to_node = {}
        self.idx_to_node = {}
        self.callpath_to_idx = {}
        self.global_nid = 0

        self.default_metric = None

        self.timer = Timer()

        if isinstance(self.filename_or_caliperreader, str):
            _, self.filename_ext = os.path.splitext(filename_or_caliperreader)

        if isinstance(self.string_attributes, str):
            self.string_attributes = [self.string_attributes]

    def read_metrics(self, ctx="path"):
        all_metrics = []
        records = self.filename_or_caliperreader.records

        # read metadata from the caliper reader
        for record in records:
            node_dict = {}
            if ctx in record:
                # only parse records that have spot.channel=regionprofile or no
                # spot.channel attribute
                if (
                    "spot.channel" in record
                    and record["spot.channel"] == "regionprofile"
                ) or "spot.channel" not in record:
                    # get the node label and callpath for the record
                    if isinstance(record[ctx], list):
                        # specify how to parse cupti records
                        if "cupti.activity.kind" in record:
                            if record["cupti.activity.kind"] == "kernel":
                                node_label = record["cupti.kernel.name"]
                                node_callpath = tuple(record[ctx] + [node_label])
                            elif record["cupti.activity.kind"] == "memcpy":
                                node_label = record["cupti.activity.kind"]
                                node_callpath = tuple(record[ctx] + [node_label])
                        else:
                            node_label = record[ctx][-1]
                            node_callpath = tuple(record[ctx])
                    else:
                        node_label = record[ctx]
                        node_callpath = tuple([record[ctx]])

                    if "spot.channel" in record:
                        node_dict["spot.channel"] = record["spot.channel"]

                    # get node nid based on callpath
                    node_dict["nid"] = self.callpath_to_idx.get(node_callpath)

                    for item in record.keys():
                        if item != ctx:
                            if (
                                self.filename_or_caliperreader.attribute(
                                    item
                                ).attribute_type()
                                == "double"
                            ):
                                node_dict[item] = float(record[item])
                                if item not in self.record_data_cols:
                                    self.record_data_cols.append(item)
                            elif (
                                self.filename_or_caliperreader.attribute(
                                    item
                                ).attribute_type()
                                == "int"
                            ):
                                node_dict[item] = int(record[item])
                                if item not in self.record_data_cols:
                                    self.record_data_cols.append(item)
                            elif (
                                self.filename_or_caliperreader.attribute(
                                    item
                                ).attribute_type()
                                == "string"
                            ):
                                if item in self.string_attributes:
                                    node_dict[item] = record[item]
                                    if item not in self.record_data_cols:
                                        self.record_data_cols.append(item)

                    all_metrics.append(node_dict)

        # make list of metric columns
        for col in self.record_data_cols:
            if self.filename_or_caliperreader.attribute(col).is_value():
                self.metric_cols.append(col)

        df_metrics = pd.DataFrame.from_dict(data=all_metrics)
        df_new = df_metrics.groupby(df_metrics["nid"]).aggregate("first").reset_index()
        return df_new

    def create_graph(self, ctx="path"):
        list_roots = []

        def _create_parent(child_node, parent_callpath):
            """We may encounter a parent node in the callpath before we see it
            as a child node. In this case, we need to create a hatchet node for
            the parent.

            This function recursively creates parent nodes in a callpath
            until it reaches the already existing parent in that callpath.
            """
            parent_node = self.callpath_to_node.get(parent_callpath)

            if parent_node:
                # return if arrives at the parent
                parent_node.add_child(child_node)
                child_node.add_parent(parent_node)
                return
            else:
                # else create the parent and add parent/child

                # if root node, end recursive call to create parent nodes
                if not parent_callpath:
                    list_roots.append(child_node)

                    node_dict = dict(
                        {
                            "name": child_node.frame["name"],
                            "node": child_node,
                            "nid": self.global_nid,
                        }
                    )

                    self.idx_to_node[self.global_nid] = node_dict
                    self.global_nid += 1
                else:
                    grandparent_callpath = parent_callpath[:-1]
                    parent_name = parent_callpath[-1]

                    parent_node = Node(
                        Frame({"type": "function", "name": parent_name}), None
                    )

                    self.callpath_to_node[parent_callpath] = parent_node
                    self.callpath_to_idx[parent_callpath] = self.global_nid

                    node_dict = dict(
                        {
                            "name": parent_name,
                            "node": parent_node,
                            "nid": self.global_nid,
                        },
                    )
                    self.idx_to_node[self.global_nid] = node_dict
                    self.global_nid += 1

                    parent_node.add_child(child_node)
                    child_node.add_parent(parent_node)
                    _create_parent(parent_node, grandparent_callpath)

        parent_hnode = None
        records = self.filename_or_caliperreader.records

        for record in records:
            node_label = ""
            if ctx in record:
                if (
                    "spot.channel" in record
                    and record["spot.channel"] == "regionprofile"
                ) or "spot.channel" not in record:
                    # if it's a list, then it's a callpath
                    if isinstance(record[ctx], list):
                        # specify how to parse cupti records
                        if "cupti.activity.kind" in record:
                            if record["cupti.activity.kind"] == "kernel":
                                node_label = record["cupti.kernel.name"]
                                node_callpath = tuple(record[ctx] + [node_label])
                                parent_callpath = node_callpath[:-1]
                                node_type = "kernel"
                            elif record["cupti.activity.kind"] == "memcpy":
                                node_label = record["cupti.activity.kind"]
                                node_callpath = tuple(record[ctx] + [node_label])
                                parent_callpath = node_callpath[:-1]
                                node_type = "memcpy"
                            else:
                                Exception("Haven't seen this activity kind yet")
                        else:
                            node_label = record[ctx][-1]
                            node_callpath = tuple(record[ctx])
                            parent_callpath = node_callpath[:-1]
                            node_type = "function"

                        hnode = self.callpath_to_node.get(node_callpath)

                        if not hnode:
                            frame = Frame({"type": node_type, "name": node_label})
                            hnode = Node(frame, None)
                            self.callpath_to_node[node_callpath] = hnode

                            # get parent from node callpath
                            parent_hnode = self.callpath_to_node.get(parent_callpath)

                            # create parent if it doesn't exist
                            # else if parent already exists, add child-parent
                            if not parent_hnode:
                                _create_parent(hnode, parent_callpath)
                            else:
                                parent_hnode.add_child(hnode)
                                hnode.add_parent(parent_hnode)

                            self.callpath_to_idx[node_callpath] = self.global_nid
                            node_dict = dict(
                                {
                                    "name": node_label,
                                    "node": hnode,
                                    "nid": self.global_nid,
                                },
                            )
                            self.idx_to_node[self.global_nid] = node_dict
                            self.global_nid += 1

                    # if it's a string, then it's a root
                    else:
                        root_label = record[ctx]
                        root_callpath = tuple([root_label])

                        if root_callpath not in self.callpath_to_node:
                            # create the root since it doesn't exist
                            frame = Frame({"type": "function", "name": root_label})
                            graph_root = Node(frame, None)

                            # store callpaths to identify the root
                            self.callpath_to_node[root_callpath] = graph_root
                            self.callpath_to_idx[root_callpath] = self.global_nid
                            list_roots.append(graph_root)

                            node_dict = dict(
                                {
                                    "name": root_label,
                                    "node": graph_root,
                                    "nid": self.global_nid,
                                }
                            )

                            self.idx_to_node[self.global_nid] = node_dict
                            self.global_nid += 1

        return list_roots

    def _parse_metadata(self, mdata):
        """Convert Caliper Metadata values into correct Python objects.

        Args:
            mdata (dict[str: str]): metadata to convert

        Return:
            (dict[str: str]): modified metadata
        """
        parsed_mdata = {}
        for k, v in mdata.items():
            # If the value is an int, convert it to an int.
            try:
                parsed_mdata[k] = int(v)
            except ValueError:
                # If the value is a float, convert it to a float
                try:
                    parsed_mdata[k] = float(v)
                except ValueError:
                    # If the value is a list or tuple, convert it to a list or
                    # tuple
                    if v.startswith("[") and v.endswith("]"):
                        parsed_mdata[k] = [
                            elem.strip() for elem in v.strip("][").split(",")
                        ]
                    elif v.startswith("(") and v.endswith(")"):
                        parsed_mdata[k] = [
                            elem.strip() for elem in v.strip(")(").split(",")
                        ]
                    # If the value is a string, just save it as-is
                    else:
                        parsed_mdata[k] = v
        return parsed_mdata

    def read(self):
        """Read the caliper records to extract the calling context tree."""
        if isinstance(self.filename_or_caliperreader, str):
            if self.filename_ext != ".cali":
                raise ValueError("from_caliperreader() needs a .cali file")
            else:
                cali_file = self.filename_or_caliperreader
                self.filename_or_caliperreader = cr.CaliperReader()
                self.filename_or_caliperreader.read(cali_file)

        with self.timer.phase("graph construction"):
            list_roots = self.create_graph()
        self.df_nodes = pd.DataFrame(data=list(self.idx_to_node.values()))

        # create a graph object once all the nodes have been added
        graph = Graph(list_roots)
        graph.enumerate_traverse()

        with self.timer.phase("read metrics"):
            df_fixed_data = self.read_metrics()

        metrics = pd.DataFrame.from_dict(data=df_fixed_data)

        # add missing intermediate nodes to the df_fixed_data dataframe
        if "mpi.rank" in df_fixed_data.columns:
            num_ranks = metrics["mpi.rank"].max() + 1
            rank_list = range(0, num_ranks)

        # create a standard dict to be used for filling all missing rows
        default_metric_dict = {}
        for idx, col in enumerate(self.record_data_cols):
            if self.filename_or_caliperreader.attribute(col).is_value():
                default_metric_dict[list(self.record_data_cols)[idx]] = 0
            else:
                default_metric_dict[list(self.record_data_cols)[idx]] = None

        # create a list of dicts, one dict for each missing row
        missing_nodes = []
        for iteridx, row in self.df_nodes.iterrows():
            # check if df_nodes row exists in df_fixed_data
            metric_rows = df_fixed_data.loc[metrics["nid"] == row["nid"]]
            if "mpi.rank" not in self.metric_cols:
                if metric_rows.empty:
                    # add a single row
                    node_dict = dict(default_metric_dict)
                    missing_nodes.append(node_dict)
            else:
                if metric_rows.empty:
                    # add a row per MPI rank
                    for rank in rank_list:
                        node_dict = dict(default_metric_dict)
                        node_dict["nid"] = row["nid"]
                        node_dict["mpi.rank"] = rank
                        missing_nodes.append(node_dict)
                elif len(metric_rows) < num_ranks:
                    # add a row for each missing MPI rank
                    present_ranks = metric_rows["mpi.rank"].values
                    missing_ranks = [x for x in rank_list if x not in present_ranks]
                    for rank in missing_ranks:
                        node_dict = dict(default_metric_dict)
                        node_dict["nid"] = row["nid"]
                        node_dict["mpi.rank"] = rank
                        missing_nodes.append(node_dict)

        df_missing = pd.DataFrame.from_dict(data=missing_nodes)
        df_metrics = pd.concat([df_fixed_data, df_missing], sort=False)

        # rename columns to user-readable metric names (i.e., aliases)
        if not self.use_native_metric_names:
            for col in df_metrics.columns:
                if col == "nid":
                    continue
                alias = self.filename_or_caliperreader.attribute(col).get(
                    "attribute.alias"
                )
                if alias:
                    # update column name in metrics dataframe
                    df_metrics.rename(columns={col: alias}, inplace=True)

                    # also update list of metric columns
                    self.metric_cols = [
                        alias if item == col else item for item in self.metric_cols
                    ]

        # dict mapping old to new column names to make columns consistent with
        # other readers
        old_to_new = {
            "mpi.rank": "rank",
            "module#cali.sampler.pc": "module",
            "sum#time.duration": "time",
            "sum#avg#sum#time.duration": "time",
            "inclusive#sum#time.duration": "time (inc)",
            "sum#avg#inclusive#sum#time.duration": "time (inc)",
        }

        # change column names
        new_cols = []
        for col in df_metrics.columns:
            if col in old_to_new:
                new_cols.append(old_to_new[col])
            else:
                new_cols.append(col)
        df_metrics.columns = new_cols

        # create list of exclusive and inclusive metric columns
        exc_metrics = []
        inc_metrics = []
        for column in self.metric_cols:
            # ignore rank as an exc or inc metric
            if column == "mpi.rank":
                continue

            # add new column names to list of metrics if inc or inclusive in
            # old column names
            if "(inc)" in column or "inclusive" in column:
                if column in old_to_new:
                    column = old_to_new[column]
                inc_metrics.append(column)
            else:
                if column in old_to_new:
                    column = old_to_new[column]
                exc_metrics.append(column)

        with self.timer.phase("data frame"):
            # merge the metrics and node dataframes on the nid column
            dataframe = pd.merge(df_metrics, self.df_nodes, on="nid")

            # set the index to be a MultiIndex
            indices = ["node"]
            if "rank" in dataframe.columns:
                indices.append("rank")
            dataframe.set_index(indices, inplace=True)
            dataframe.sort_index(inplace=True)

        # set the default metric
        if self.default_metric is None:
            if "time (inc)" in dataframe.columns:
                self.default_metric = "time"
            elif "avg#inclusive#sum#time.duration" in dataframe.columns:
                self.default_metric = "avg#inclusive#sum#time.duration"
            elif len(inc_metrics) > 0:
                self.default_metric = inc_metrics[0]
            elif len(exc_metrics) > 0:
                self.default_metric = exc_metrics[0]

        metadata = self.filename_or_caliperreader.globals
        parsed_metadata = self._parse_metadata(metadata)

        return hatchet.graphframe.GraphFrame(
            graph,
            dataframe,
            exc_metrics,
            inc_metrics,
            self.default_metric,
            metadata=parsed_metadata,
        )



# Copyright 2017-2023 Lawrence Livermore National Security, LLC and other
# Hatchet Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

import cython
import numpy as np
from libc.math cimport floor


# Description: All cython code related to graphframe.py
@cython.boundscheck(False)
def insert_one_for_self_nodes(const long snio_len, short[:] self_missing_node, const long[:] snio_indices):
    """Adds a '1' where rows are in self but not in other."""
    for i in range(snio_len):
        self_missing_node[snio_indices[i]] = 1

cpdef fast_not_isin(const unsigned long long[:,:] arr1, const unsigned long long[:,:] arr2, const long arr1_len, const long arr2_len):
    """Check if elements in arr1 exist in arr2.

    A fast check to see if each element in arr1 exists in arr2. This returns a
    vector mask equivalent to what the operation ~df.isin( ... ) would return.

    Arguments:
        arr1 (unsigned long long [][]): The array of values we are searching for.
        arr2 (unsigned long long [][]): The sorted array of values we are searching in.

    Returns:
        (bool []): A boolean mask over arr1 indicating whether each element is or is not in
            the function. True indicates that an element is not in arr2.
    """
    result = np.zeros(len(arr1), dtype=np.bool_)
    cdef long index = -1
    cdef unsigned long long prior = -1
    cdef long i = 0

    for i in range(arr1_len):
        # mini optimization: don't perform binsearch if we already know node is/isnot in graph
        if prior == arr1[i][0]:
            result[arr1[i][1]] = result[arr1[i-1][1]]
        else:
            index = binary_search(arr2, arr2_len, arr1[i][0])
            if index == -1:
                result[arr1[i][1]] = True
            else:
                result[arr1[i][1]] = False

    prior = arr1[i][0]

    return result

cdef binary_search(const unsigned long long[:,:] array, const long arr_len, const unsigned long long key):
    """Basic Binary Search.

    Arguments:
        array (unsigned long long [][]): Array we are searching in (must be sorted).

    Returns: The index of key if found, -1 if not found.
    """
    cdef long L = 0
    cdef long R = arr_len - 1
    cdef long midpt = 0
    while (L <= R):
        midpt = <long>floor((L+R)/2)
        if array[midpt][0] < key:
            L = midpt + 1
        elif array[midpt][0] > key:
            R = midpt - 1
        else:
            return midpt
    return -1



# Copyright 2017-2023 Lawrence Livermore National Security, LLC and other
# Hatchet Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

from hatchet import GraphFrame


def test_read_json(json_graphframe_specification):
    jgs = ""
    with open(json_graphframe_specification, "r") as f:
        jgs = f.read()
    gf = GraphFrame.from_json(jgs)

    assert len(gf.dataframe) == 24
    assert len(gf.graph) == 24
    assert gf.graph.roots[0].frame["name"] == "foo"


def test_write_json(json_graphframe_specification):
    jgs = ""
    with open(json_graphframe_specification, "r") as f:
        jgs = f.read()
    gf = GraphFrame.from_json(jgs)
    json_out = gf.to_json()

    assert "".join(sorted("".join(sorted(jgs.split())))) == "".join(
        sorted("".join(json_out.split()))
    )
